#include "kernel_interface.h"
#include "zx_adapter.h"
#include "context.h"
#include "global.h"
#include "powermgr.h"
#include "vidmm.h"
#include "vidsch.h"
#include "perfevent.h"

krnl_import_func_list_t *zx = NULL;

static int krnl_create_device(void *data, void *filp, unsigned int *hDevice);
static int krnl_create_context(void *data, zx_create_context_t *create_context, enum zx_mem_space mem_space);
static void krnl_destroy_device(void *data, unsigned int hDevice);

static int krnl_cil2_misc(void *data, zx_cil2_misc_t* zx_misc)
{
    adapter_t *adapter = data;
    krnl_cil2_misc_t misc;

    misc.zx_misc = zx_misc;
    misc.context = get_from_handle(&adapter->hdl_mgr, zx_misc->context);
    misc.device = get_from_handle(&adapter->hdl_mgr, zx_misc->device);

    return vidsch_cil2_misc(misc.device, &misc);
}

static void* krnl_pre_init_adapter(void *pdev, krnl_adapter_init_info_t *info, krnl_import_func_list_t *import)
{
    adapter_t *adapter     = NULL;
    platform_caps_t caps   = {0};

    zx = import;

    util_init_log();

    adapter = zx_calloc(sizeof(adapter_t));

    if(adapter == NULL)
    {
        zx_error("allocate adapter fail!\n");

        return NULL;
    }

    adapter->paging_lock        = zx_create_mutex();
    adapter->device_list_lock   = zx_create_mutex();
    adapter->gart_table_lock    = zx_create_mutex();
    adapter->lock               = zx_create_spinlock(0);
    adapter->log_fd             = zx_logmsg_init();

    adapter->vcp_index_cnt[VCP0_INDEX] = 0;
    adapter->vcp_index_cnt[VCP1_INDEX] = 0;
    adapter->start_index = MIN_VIDEO_SEQ_INDEX;
    adapter->end_index = MAX_VIDEO_SEQ_INDEX_ADD1;
    zx_memset((void*)(adapter->bVideoSeqIndex), 0, sizeof(adapter->bVideoSeqIndex));

    list_init_head(&adapter->device_list);
    list_init_head(&adapter->sync_obj_list);

    adapter->index = info->minor_index;
    adapter->os_device.pdev = pdev;

    adapter->ctl_flags.worker_thread_enable = TRUE;
    adapter->ctl_flags.split_enable         = TRUE;
    adapter->hw_caps.fence_interrupt_enable = TRUE;
    adapter->ctl_flags.recovery_enable        = FALSE;
    adapter->ctl_flags.dump_hang_info_enable  = FALSE;
    adapter->ctl_flags.dump_hang_info_to_file = FALSE;
    adapter->ctl_flags.swap_enable            = FALSE;
    adapter->ctl_flags.perf_event_enable      = TRUE;
    adapter->ctl_flags.local_for_display_only = TRUE;

    zx_query_platform_caps(pdev, &caps);

    adapter->sys_caps.iommu_enabled  = caps.iommu_support;
    if(adapter->ctl_flags.run_in_guest)
    {
        zx_info("detect running in guest mode, force iommu enabled.\n");
        adapter->sys_caps.iommu_enabled = TRUE;
    }

    adapter->hw_caps.address_mode_64bit = !caps.system_need_dma32;
    adapter->os_page_size  = caps.page_size;
    adapter->os_page_shift = caps.page_shift;

    glb_get_pci_config_info(adapter);

    glb_init_chip_id(adapter,info);

    handle_mgr_create(&adapter->hdl_mgr);

    return  adapter;
}

static  void krnl_init_adapter(void* adp, int reserved_vmem, void *disp_info)
{
    adapter_t * adapter     = (adapter_t*)adp;

    adapter->disp_info = disp_info;

    vidmm_init(adapter, reserved_vmem);

    vidsch_create(adapter);

    zx_register_trace_events();

    perf_event_init(adapter);

    //zx_info("adapter->ctl_flags.worker_thread_enable %x\n", adapter->ctl_flags.worker_thread_enable);

    zx_info("sys caps: os page size :0x%x, os page shift:0x%x, iommu_support:%x\n",
            adapter->os_page_size, adapter->os_page_shift,
            adapter->sys_caps.iommu_enabled);

    zx_info("hw cfg: chip slice mask :%x, miu channel index:%x, miu channel size index:%x\n",
            adapter->hw_caps.chip_slice_mask, adapter->hw_caps.miu_channel_num,
            adapter->hw_caps.miu_channel_size);

    zx_info("hw caps: secure:%d, snoop:%d, 4K_page:%d, 64K_page:%d, Paging:%d, CG:%d, DFS:%d, nonsnoop_patch:%d\n",
         adapter->hw_caps.secure_range_enable, adapter->hw_caps.support_snooping,
         adapter->hw_caps.page_4k_enable, adapter->hw_caps.page_64k_enable, adapter->ctl_flags.paging_enable,
         adapter->pwm_level.EnableClockGating, adapter->hw_caps.dvfs_enable, adapter->hw_caps.nonsnoop_patch_enable);

    zx_info("power caps: ClockGating:%d, PowerGating:%d, pwm manaul mode %x\n",
        adapter->pwm_level.EnableClockGating, adapter->pwm_level.EnablePowerGating, adapter->pm_caps.pwm_manual);

    zx_info("power caps cg auto mode gfx %d, vcp %d, vpp %d, cg manual mode gfx %d, vcp %d, vpp %d \n",
        adapter->pm_caps.gfx_cg_auto, adapter->pm_caps.vcp_cg_auto, adapter->pm_caps.vpp_cg_auto,
        adapter->pm_caps.gfx_cg_manual, adapter->pm_caps.vcp_cg_manual, adapter->pm_caps.vpp_cg_manual);

    zx_info("power caps pg auto mode gfx %d, vcp %d, vpp %d, pg manual mode gfx %d, vcp %d, vpp %d \n",
        adapter->pm_caps.gfx_pg_auto, adapter->pm_caps.vcp_pg_auto, adapter->pm_caps.vpp_pg_auto,
        adapter->pm_caps.gfx_pg_manual, adapter->pm_caps.vcp_pg_manual, adapter->pm_caps.vpp_pg_manual);

    zx_info("power caps dvfs auto mode gfx %d, vcp %d, vpp %d, dvfs manual mode gfx %d, vcp %d, vpp %d \n",
        adapter->pm_caps.gfx_dvfs_auto, adapter->pm_caps.vcp_dvfs_auto, adapter->pm_caps.vpp_dvfs_auto,
        adapter->pm_caps.gfx_dvfs_force, adapter->pm_caps.vcp_dvfs_force, adapter->pm_caps.vpp_dvfs_force);

    zx_info("power caps dvfs mem auto mode gfx %d, vcp %d, vpp %d, dvfs output mode gfx %d, vcp %d, vpp %d \n",
        adapter->pm_caps.gfx_mem_dvfs, adapter->pm_caps.vcp_mem_dvfs, adapter->pm_caps.vpp_mem_dvfs,
        adapter->pm_caps.gfx_output_mode, adapter->pm_caps.vcp_output_mode, adapter->pm_caps.vpp_output_mode);

    zx_info("Ctrl: Recovery:%d, WK-thread:%d, Hang-Dump:%d, IOMMU:%d, RunOnQT:%d, RunInGuest:%d, PwmMode:0x%x, NonsnoopEnable:%d\n",
        adapter->ctl_flags.recovery_enable, adapter->ctl_flags.worker_thread_enable,
        adapter->ctl_flags.hang_dump, adapter->sys_caps.iommu_enabled,
        adapter->ctl_flags.run_on_qt, adapter->ctl_flags.run_in_guest,
        adapter->pm_caps.pwm_mode, adapter->hw_caps.snoop_only ? 0 : 1);

}

static void krnl_deinit_adapter(void *data)
{
    adapter_t *adapter = data;

    perf_event_deinit(adapter);

    zx_unregister_trace_events();

    cm_destroy_remained_device(adapter);

    vidsch_destroy(adapter);

    vidmm_destroy(adapter);

    handle_mgr_destroy(&adapter->hdl_mgr);

    zx_destroy_mutex(adapter->gart_table_lock);

    zx_destroy_mutex(adapter->paging_lock);

    zx_destroy_mutex(adapter->device_list_lock);

    zx_destroy_spinlock(adapter->lock);

    glb_fini_bus_config(adapter);

    zx_logmsg_deinit(adapter->log_fd);

    zx_free(adapter);
}


static  void krnl_get_adapter_info(void*  adp, adapter_info_t*  adapter_info)
{
    adapter_t *adapter = (adapter_t *)adp;
    if(adapter && adapter_info)
    {
        adapter_info->family_id = adapter->family_id;
        adapter_info->generic_id = adapter->generic_id;
        adapter_info->chip_id = adapter->chip_id;
        adapter_info->mmio  = adapter->mmio;
        adapter_info->mmio_size = adapter->mmio_vma->size;
        adapter_info->primary = adapter->primary;
        adapter_info->fb_bus_addr = adapter->vidmm_bus_addr;
        adapter_info->fb_total_size = adapter->Visible_vram_size;
        adapter_info->run_on_qt = adapter->ctl_flags.run_on_qt;
        adapter_info->gpio      = adapter->gpio;
        adapter_info->patch_fence_intr_lost = (adapter->hw_patch_mask0 & PATCH_FENCE_INTERRUPT_LOST) ? 1 : 0;
    }
}

static  void  krnl_update_adapter_info(void* adp, adapter_info_t*  adapter_info)
{
    adapter_t *adapter = (adapter_t *)adp;
    if(adapter && adapter_info)
    {
        adapter->hw_caps.snoop_only = (adapter_info->snoop_only)? 1 : 0;
        adapter->low_top_address = adapter_info->low_top_addr;
        adapter->hw_caps.ta_enable = adapter_info->ta_enable;
        if (adapter_info->run_on_hypervisor)
        {
            adapter->ctl_flags.run_in_guest = 1;
        }

        if(adapter->chip_id == CHIP_CHX004)
        {
            adapter_info->chan_num = 2;

            if(adapter->low_top_address == 0){
                zx_warning("Fix up: get low_top_address 0, hard code to 0xc0000000\n");
                adapter->low_top_address = 0xc0000000;
            }
        }

        adapter->hw_caps.miu_channel_num = adapter_info->chan_num;

        if(adapter->Real_vram_size == 0)
        {
            adapter->Real_vram_size = adapter_info->total_mem_size_mb;
            adapter->Real_vram_size <<= 20;
        }
        else if((adapter->Real_vram_size>>20) > adapter_info->total_mem_size_mb)
        {
            zx_info("use zx_local_size as local memory, Real_vram_size:%lldM, total_mem_size_mb:%lldM\n", (adapter->Real_vram_size>>20), adapter_info->total_mem_size_mb);
            zx_warning("set local memory out range actual memory size, force adjust actual memory");
            adapter->Real_vram_size = adapter_info->total_mem_size_mb;
            adapter->Real_vram_size <<= 20;
        }
        else
        {
            zx_info("use params zx_local_size_g to set local memory, not total_mem_size_mb\n");
        }

        adapter->UnAval_vram_size = (adapter_info->total_mem_size_mb - adapter_info->avai_mem_size_mb);
        adapter->UnAval_vram_size <<= 20;

        zx_info("video memory size: %lldM.\n", adapter->Real_vram_size >> 20);
        zx_info("video unav size:   %lldM.\n", adapter->UnAval_vram_size >> 20);

#if 0
        //slice mask setting from kmd module para, for debug use.
        //if(adapter_info->chip_slice_mask)
        //{
        //    adapter->hw_caps.chip_slice_mask    =  adapter_info->chip_slice_mask;
        //}
        //if hang dump , force set slice mask is 0x01
        if (adapter->ctl_flags.hang_dump)
        {
            adapter->hw_caps.chip_slice_mask = 0x01;
        }
#endif
        zx_info("krnl_update_adapter_info chip_slice_mask: 0x%x\n", adapter->hw_caps.chip_slice_mask);
    }
}

static void krnl_dump_resource(struct os_printer *p, void *data, int dump_index, int iga_index)
{
    adapter_t *adapter = data;

    switch(dump_index)
    {
       case 0:
           util_print_log();
           break;
       case 1:
           vidmm_dump_resource(adapter);
           break;
       case 2:
           vidsch_test_memory(adapter);
           break;
       case 3:
           vidsch_dump(p, adapter);
           break;
       case 4:
           cm_dump_resource(adapter);
           break;
       case 5:
           util_print_log();
           vidmm_dump_resource(adapter);
           vidsch_dump(p,adapter);
           cm_dump_resource(adapter);
           break;
       case 0x7000:
           vidsch_force_wakup(adapter);
           break;
       default:
           break;
    }
}

static void krnl_debugfs_dump(struct os_seq_file *seq_file, void *data, int type, void* arg)
{
    adapter_t *adapter = data;
    switch(type)
    {
        case DEBUGFS_NODE_DEVICE:
        {
            unsigned int handle = *((unsigned int*)arg);

            gpu_device_t *device  = get_from_handle(&adapter->hdl_mgr, handle);

            if(device != NULL)
            {
                cm_dump_device_alloctions(seq_file, adapter, device);
            }
            break;
        }

        case DEBUGFS_NODE_HEAP:
        {
            int id = *((int*)arg);

            vidmm_dump_heap(seq_file, adapter, id);
            break;
        }
        case DEBUGFS_NODE_INFO:
            vidsch_dump_info(seq_file, adapter);
            break;

        case DEBUGFS_NODE_MEMTRACK:
        {
            int pid = *((int *)arg);
            vidmm_dump_memtrack(seq_file, adapter, pid);
            break;
        }

        case DEBUGFS_NODE_VIDSCH:
        {
            struct os_printer *p = (struct os_printer *) arg;
            vidsch_dump(p, adapter);
            break;
        }

        case DEBUGFS_NODE_DEBUGBUS:
        {
            struct os_printer *p = (struct os_printer *) arg;
            vidsch_dump_debugbus(p, adapter);
            break;
        }

        case DEBUGFS_NODE_CPLL:
        {
            long long readVal1,readVal2,tmp;

            adapter_t *adapter=data;

            readVal1=zx_read32(adapter->mmio+0x300C0);
            readVal1=readVal1<<32;
            tmp=zx_read32(adapter->mmio+0x300BC);
            readVal1+=tmp;

            zx_msleep(5000);

            readVal2=zx_read32(adapter->mmio+0x300C0);
            readVal2=readVal2<<32;
            tmp=zx_read32(adapter->mmio+0x300BC);
            readVal2+=tmp;

            zx_seq_printf(seq_file, "CMU frequency: %dMHz\n", (readVal2 - readVal1) / 5 / 1000000);
            break;
        }
        case DEBUGFS_NODE_TIMER_TIMEOUT:
        {
            zx_seq_printf(seq_file, "timeout of the timer is %d ms\n", adapter->timeout_ms);
            break;   
        }
        default:
            break;
    }
}


static int krnl_create_device(void *data, void *filp, unsigned int *hDevice)
{
    adapter_t    *adapter = data;
    gpu_device_t *device;

    int status = E_INVALIDARG;

    device = cm_create_device(adapter, filp);

    if((device != NULL) && (hDevice != NULL))
    {
        *hDevice = device->handle;
        device->video_seq_index = 0xFFFF;

        status = S_OK;
    }

    return status;
}

static void krnl_destroy_device(void *data, unsigned int hDevice)
{
    adapter_t    *adapter = data;
    gpu_device_t *device  = get_from_handle(&adapter->hdl_mgr, hDevice);

    if(device != NULL)
    {
        if(device->video_seq_index != 0xFFFF)
        {
            adapter->bVideoSeqIndex[device->video_seq_index] = 0;
            device->video_seq_index = 0xFFFF;
        }
        cm_destroy_device(adapter, device);
    }
    else
    {
        zx_error("%s: invalid device handle %x.\n", __func__, hDevice);
    }

}

static void krnl_final_cleanup(void *data, unsigned int gpu_device)
{
    adapter_t    *adapter = data;
    gpu_device_t *device  = get_from_handle(&adapter->hdl_mgr, gpu_device);
    if(device != NULL)
    {

        if(device->video_seq_index != 0xFFFF)
        {
            adapter->bVideoSeqIndex[device->video_seq_index] = 0;
            device->video_seq_index = 0xFFFF;
        }
    }
    cm_destroy_device(adapter, device);
}

static void krnl_wait_chip_idle(void *data)
{
    adapter_t *adapter = data;

    vidsch_wait_chip_idle(adapter, ALL_ENGINE_MASK);
}

static int krnl_create_allocation(void *data, zx_create_allocation_t *create_data, void *bo)
{
    adapter_t     *adapter = data;
    gpu_device_t  *device  = get_from_handle(&adapter->hdl_mgr, create_data->device);
    vidmm_create_allocation_arg_t create = {0};
    vidmm_rename_create_t rename_create = {0,};
    vidmm_zx_create_t zx_create = {0, };
    int result = S_OK;

    if(device == NULL)
    {
        zx_error("%s: invalid device handle %x.\n", __func__, create_data->device);

        result = E_INVALIDARG;

        return result;
    }
    if (create_data->reference)
    {
        vidmm_allocation_t  *reference = get_from_handle(&adapter->hdl_mgr, create_data->reference);

        rename_create.reference     = reference;
        create.allocation_count     = 1;
        create.create_type          = VIDMM_CREATE_TYPE_RENAME;
        create.rename_list          = &rename_create;
    }
    else
    {
        zx_create.create_data       = create_data;
        create.allocation_count     = 1;
        create.create_type          = VIDMM_CREATE_TYPE_ZX;
        create.create_list          = &zx_create;
    }

    create.bos          = &bo;

    result = vidmm_create_allocation(device, &create);

    if (create.create_type == VIDMM_CREATE_TYPE_RENAME)
    {
        create_data->size       = rename_create.Size;
        create_data->allocation = rename_create.hAllocation;
    }

    return result;
}

static int krnl_create_allocation_from_pages(void *data, zx_create_allocation_t *create_data, struct os_pages_memory *pages, void *bo)
{
    adapter_t     *adapter = data;
    gpu_device_t  *device  = get_from_handle(&adapter->hdl_mgr, create_data->device);
    vidmm_create_allocation_arg_t create = {0};
    vidmm_zx_create_t zx_create = {0, };
    int result = S_OK;

    if(device == NULL)
    {
        zx_error("%s: invalid device handle %x.\n", __func__, create_data->device);

        result = E_INVALIDARG;

        return result;
    }

    zx_create.create_data       = create_data;
    zx_create.import_pages_mem  = pages;

    create.create_type          = VIDMM_CREATE_TYPE_ZX;
    create.allocation_count     = 1;
    create.create_list          = &zx_create;
    create.bos                  = &bo;

    result = vidmm_create_allocation(device, &create);

    return result;
}

static int krnl_create_allocation_list(void *data, zx_create_resource_t *create_data, void **bos)
{
    adapter_t        *adapter = data;
    gpu_device_t     *device  = get_from_handle(&adapter->hdl_mgr, create_data->device);
    vidmm_create_allocation_arg_t create = {0};

    create.create_type      = VIDMM_CREATE_TYPE_ESCAPE;
    create.allocation_count = create_data->NumAllocations;
    create.info_list        = ptr64_to_ptr(create_data->pAllocationInfo);
    create.bos = bos;

    return vidmm_create_allocation(device, &create);
}

static void krnl_destroy_allocation(void *data, zx_destroy_allocation_t *destroy_data)
{
    adapter_t          *adapter    = data;
    gpu_device_t       *device     = get_from_handle(&adapter->hdl_mgr, destroy_data->device);
    vidmm_allocation_t *allocation = get_from_handle(&adapter->hdl_mgr, destroy_data->allocation);

    vidmm_destroy_allocatin_arg_t destroy = {0};

    if (device == NULL && allocation != NULL)
    {
        device = allocation->device;
    }

    if(device == NULL)
    {
        zx_error("%s: invalid device handle %x.\n", __func__, destroy_data->device);

        return;
    }

    if(allocation == NULL)
    {
        zx_error("%s: invalid allocation handle %x.\n", __func__, destroy_data->allocation);

        return;
    }

    destroy.allocation_count = 1;
    destroy.allocation_list  = &allocation;

    vidmm_destroy_allocation(device, &destroy);
}

static void krnl_get_map_allocation_info(void *data, unsigned int hAllocation, zx_map_argu_t *map)
{
    adapter_t          *adapter = data;
    vidmm_allocation_t *allocation = get_from_handle(&adapter->hdl_mgr, hAllocation);

    return vidmm_get_map_allocation_info(adapter, allocation, map);
}

static void krnl_wait_allocation_idle(void *data, zx_wait_allocation_idle_t *wait_allocation)
{
    adapter_t          *adapter    = data;
    vidmm_allocation_t *allocation = get_from_handle(&adapter->hdl_mgr, wait_allocation->hAllocation);

    zx_begin_section_trace_event("wait_allocation_idle");
    zx_counter_trace_event("arg_allocation", wait_allocation->hAllocation);

    vidsch_wait_allocation_idle(adapter, wait_allocation->engine_mask, 0, allocation);

    zx_end_section_trace_event(0);
}

static int krnl_save_state(void *data, int need_save_memory)
{
    adapter_t *adapter = data;
    return  pm_save_state(adapter, need_save_memory);
}

static int krnl_restore_state(void *data)
{
    adapter_t *adapter = data;
    int        result  = S_OK;

    result = pm_restore_state(adapter);

    return result;
}

static int krnl_query_info(void* data, zx_query_info_t *info)
{
    adapter_t     *adapter = data;
    int           status   = 0;
    int           search_num = 0;
    gpu_device_t *device  = NULL;

    switch (info->type)
    {
    case ZX_QUERY_TOTAL_VRAM_SIZE:
    case ZX_QUERY_RESERV_VRAM_SIZE:
    case ZX_QUERY_CPU_VISIBLE_VRAM_SIZE:
    case ZX_QUERY_LOCAL_VRAM_TYPE:
    case ZX_QUERY_HEIGHT_ALIGN:
    case ZX_QUERY_SEGMENT_FREE_SIZE:
    case ZX_QUERY_ALLOCATION_INFO:
    case ZX_QUERY_ALLOCATION_INFO_KMD:
    case ZX_QUERY_LOCAL_ALLOCATION_MAX_SIZE:
    case ZX_QUERY_UNSNOOPABLE_INFO:
        status = vidmm_query_info(adapter, info);
        break;

    case ZX_QUERY_HW_HANG:
    case ZX_QUERY_PENDING_FRAME_NUM:
    case ZX_QUERY_GET_VIDEO_BRIDGE_BUFFER:
    case ZX_QUERY_GPU_TIME_STAMP:
    case ZX_QUERY_REGISTER_U32:
	case ZX_SET_REGISTER_U32:
        status = vidsch_query_info(adapter, info);
        break;

    case ZX_QUERY_PAGE_SWIZZLE_SUPPORT:
        info->value = 0;
        break;

    case ZX_QUERY_CHIP_ID:
        info->value = adapter->chip_id;
        break;

    case ZX_QUERY_VENDOR_ID:
        info->value = adapter->bus_config.vendor_id;
        break;

    case ZX_QUERY_DEVICE_ID:
        info->value = adapter->bus_config.device_id;
        break;

    case ZX_QUERY_REVISION_ID:
        info->value = adapter->bus_config.revision_id;
        break;

    case ZX_QUERY_CHIP_SLICE_MASK:
        info->value = adapter->hw_caps.chip_slice_mask;
        break;

    case ZX_QUERY_SECURED_ON:
        info->value = adapter->sys_caps.secure_on;
        break;

    case ZX_QUERY_VCP_INDEX:
        if (info->argu == 0xec)
        {   //encoder or AVS2
            info->value = 0;        //core id must be 0
        }
        else {
            info->value = adapter->vcp_index_cnt[VCP0_INDEX] >= adapter->vcp_index_cnt[VCP1_INDEX] ? VCP1_INDEX : VCP0_INDEX;
        }
        adapter->vcp_index_cnt[info->value]++;
        break;

    case ZX_QUERY_RET_VCP_INDEX:
        if (info->argu > VCP1_INDEX || info->argu < VCP0_INDEX)
        {
            zx_error("unknown arg (%d) for ZX_QUERY_RET_VCP_INDEX\n", info->argu);
            status = -1;
        }
        else {
            adapter->vcp_index_cnt[info->argu]--;
        }
        break;

    case ZX_QUERY_VIDEO_SEQ_INDEX:
        while(adapter->start_index < adapter->end_index)
        {
            if(adapter->bVideoSeqIndex[adapter->start_index] == 0)
            {
                info->value = adapter->start_index;
                adapter->bVideoSeqIndex[adapter->start_index] = 1;
                device = get_from_handle(&adapter->hdl_mgr, info->argu);
                if(device)
                {
                    device->video_seq_index = adapter->start_index;
                    adapter->start_index = (adapter->start_index + 1) % MAX_VIDEO_SEQ_INDEX_ADD1;
                    adapter->end_index = MAX_VIDEO_SEQ_INDEX_ADD1;
                }
                status = 0;
                break;
            }

            adapter->start_index++;

            if(adapter->start_index == adapter->end_index)
            {
                search_num++;
                adapter->end_index  = adapter->start_index;
                adapter->start_index = MIN_VIDEO_SEQ_INDEX;
            }

            if(search_num == 2)
            {
               status = -1;
               break;
            }
        }
        break;

    case ZX_QUERY_ADAPTER_INFO:
        {
            adapter_t           *adapter   = data;
            zx_adapter_info_t   adapter_info = {0};
            bus_config_t        *bus_info  = &adapter->bus_config;

            adapter_info.bus_config.DeviceID    = bus_info->device_id;
            adapter_info.bus_config.VendorID    = bus_info->vendor_id;
            adapter_info.bus_config.Command     = bus_info->command;
            adapter_info.bus_config.Status      = bus_info->status;
            adapter_info.bus_config.RevisionID  = bus_info->revision_id;
            adapter_info.bus_config.ProgIf      = bus_info->prog_if;
            adapter_info.bus_config.SubClass    = bus_info->sub_class;
            adapter_info.bus_config.BaseClass   = bus_info->base_class;
            adapter_info.bus_config.LatencyTimer= bus_info->latency_timer;
            adapter_info.bus_config.HeaderType  = bus_info->header_type;
            adapter_info.bus_config.BIST        = bus_info->bist;
            adapter_info.bus_config.LinkStatus  = bus_info->link_status;

            adapter_info.bus_config.ulBaseAddresses[0] = 0;
            adapter_info.bus_config.ulBaseAddresses[1] = 0;
            adapter_info.bus_config.ulBaseAddresses[2] = 0;
            adapter_info.bus_config.ulBaseAddresses[3] = 0;
            adapter_info.bus_config.ulBaseAddresses[4] = 0;
            adapter_info.bus_config.ulBaseAddresses[5] = 0;

            adapter_info.bus_config.CacheLineSize      = bus_info->cache_line_size;
            adapter_info.bus_config.SubsystemVendorID  = bus_info->sub_sys_vendor_id;

            adapter_info.bus_config.memorySize         = adapter->Real_vram_size - adapter->UnAval_vram_size;
            adapter_info.chipslicemask                 = adapter->hw_caps.chip_slice_mask;
            adapter_info.gpucount                      = 1;
            adapter_info.osversion                     = 0;
            adapter_info.bVideoOnly                    = adapter->hw_caps.video_only;
            adapter_info.bCTEDumpEnable                = 0;
            adapter_info.bSnoopOnly                    = adapter->hw_caps.snoop_only;

            adapter_info.segmentMapTable[0]            = 0;
            adapter_info.segmentMapTable[1]            = 1;//fb low
            adapter_info.segmentMapTable[4]            = 2;//non-snoop
            adapter_info.segmentMapTable[3]            = 3;//snoop
            adapter_info.segmentMapTable[2]            = 4;//fb high
            adapter_info.bl_slot_size                  = BL_SLOT_SIZE;
            zx_copy_to_user(info->buf, &adapter_info, sizeof(zx_adapter_info_t));
        }
        break;

    case ZX_QUERY_ACTIVE_ENGINE_COUNT:
        info->value = adapter->active_engine_count;
        break;

    case ZX_QUERY_RET_VIDEO_SEQ_INDEX:
        if(info->argu >= MIN_VIDEO_SEQ_INDEX && (info->argu < MAX_VIDEO_SEQ_INDEX_ADD1))
        {
            adapter->bVideoSeqIndex[info->argu] = 0;
        }
        break;

    case ZX_QUERY_VCP_INFO:
        {
            int index = (info->argu << 16) >> 16;
            int op    = (info->argu >> 16);

            if(index < 0 || index >= VCP_INFO_COUNT) {
                zx_error("invalid index-%d\n", index);
                return -1;
            }

            switch(op) {
                case 0:
                {
                    int i;
                    for(i=0; i<VCP_INFO_COUNT; i++) {
                        if(adapter->vcp_info[i].enable == 0) {
                            zx_copy_from_user(&adapter->vcp_info[i], info->buf, sizeof(zx_vcp_info));
                            adapter->vcp_info[i].enable = 1;
                            index = i;
                            break;
                        }
                    }

                    break;
                }
                case 1:
                    adapter->vcp_info[index].enable = 0;
                    adapter->vcp_info[index].pid = 0;
                    break;
                case 2:
                    zx_copy_to_user(info->buf, &adapter->vcp_info[index], sizeof(zx_vcp_info));
                    break;
                case 4:
                {
                    int i;
                    for(i=0; i<VCP_INFO_COUNT; i++) {
                        adapter->vcp_info[i].enable = 0;
                        adapter->vcp_info[i].pid = 0;
                    }
                    break;
                }
                case 3:
                default:
                    zx_copy_from_user(&adapter->vcp_info[index], info->buf, sizeof(zx_vcp_info));
                    break;
            }

            info->argu = index;
        }
        break;

    default:
        zx_assert(0, "unknown query type");
        status = -1;
        break;
    }

    return status;
}

static int krnl_set_callback_func(void *data, int type, void *func, void *argu)
{
    adapter_t     *adapter = data;
    int           status   = S_OK;

    switch (type)
    {
    case OS_CALLBACK_POST_EVENT:
        adapter->post_event_argu = argu;
        adapter->post_event      = func;
        break;
    case OS_CALLBACK_POST_SYNCOBJ_EVENT:
        adapter->post_sync_event_argu = argu;
        adapter->post_sync_event      = func;
        break;
    case OS_CALLBACK_DRM_CB:
        adapter->drm_cb = func;
        adapter->drm_cb_argu = argu;
        break;
    default:
        zx_error("unknown callback func type: %d.\n", type);
        status = E_INVALIDARG;
        break;
    }

    return status;
}

static int krnl_begin_perf_event(void *data, zx_begin_perf_event_t *begin_perf_event)
{
    adapter_t    *adapter = data;
    int result = 0;

    result = perf_event_begin(adapter, begin_perf_event);

    return result;
}

static int krnl_begin_miu_dump_perf_event(void *data, zx_begin_miu_dump_perf_event_t *begin_miu_perf_event)
{
    adapter_t *adapter = data;
    int result = 0;

    result = perf_event_begin_miu_dump(adapter, begin_miu_perf_event);

    return result;
}

static int krnl_end_perf_event(void *data, zx_end_perf_event_t *end_perf_event)
{
    adapter_t    *adapter = data;
    int result = 0;

    result = perf_event_end(adapter, end_perf_event);

    return result;
}

static int krnl_end_miu_dump_perf_event(void *data, zx_end_miu_dump_perf_event_t *end_miu_perf_event)
{
    adapter_t *adapter = data;
    int result = 0;

    result = perf_event_end_miu_dump(adapter, end_miu_perf_event);

    return result;
}

static int krnl_get_miu_dump_perf_event(void *data, zx_get_miu_dump_perf_event_t *get_miu_dump)
{
    adapter_t *adapter = data;
    int result = 0;

    result = perf_event_get_miu_event(adapter, get_miu_dump);

    return result;
}

static int krnl_direct_get_miu_dump_perf_event(void *data, zx_direct_get_miu_dump_perf_event_t *direct_get_dump)
{
    adapter_t *adapter = data;
    int result = 0;

    result = perf_event_direct_get_miu_dump_event(adapter, direct_get_dump);

    return result;
}

static int krnl_get_perf_event(void *data, zx_get_perf_event_t *get_event)
{
    adapter_t    *adapter = data;
    int result = 0;

    result = perf_event_get_event(adapter, get_event);

    return result;
}

static int krnl_set_miu_reg_list_perf_event(void *data, zx_miu_reg_list_perf_event_t *miu_reg_list)
{
    adapter_t *adapter = data;
    int result = 0;

    result = perf_event_set_miu_reg_list(adapter, miu_reg_list);

    return result;
}

static int krnl_send_perf_event(void *data, zx_perf_event_t *perf_event)
{
    adapter_t    *adapter = data;
    int result = 0;

    result = perf_event_add_event(adapter, perf_event);

    return result;
}

static int krnl_get_perf_status(void *data, zx_perf_status_t *perf_status)
{
    adapter_t   *adapter = data;
    int result = 0;

    result = perf_event_get_status(adapter, perf_status);

    return result;
}

static int krnl_create_context(void *data, zx_create_context_t *create_context, enum zx_mem_space mem_space)
{
    adapter_t        *adapter = data;
    gpu_device_t     *device  = get_from_handle(&adapter->hdl_mgr, create_context->device);
    gpu_context_t    *context = NULL;
    create_context_t create   = {0};
    int is_kernel = (mem_space == ZX_MEM_KERNEL) ? 1 : 0;

    //multi-engine
    create.node_ordinal    = create_context->engine_index;
    create.flags           = create_context->flags;

    context = cm_create_context(device, &create, is_kernel);

    if(context)
    {
        create_context->context                  = context->handle;
        return S_OK;
    }

    return E_FAIL;
}

static int krnl_destroy_context(void *data, zx_destroy_context_t *destroy_context)
{
    adapter_t     *adapter = data;
    gpu_device_t  *device  = get_from_handle(&adapter->hdl_mgr, destroy_context->device);
    gpu_context_t *context = get_from_handle(&adapter->hdl_mgr, destroy_context->context);

    cm_destroy_context(device, context);

    return S_OK;
}

static int krnl_wait_context_idle(void * data, unsigned int context)
{
    adapter_t *adapter         = data;
    gpu_context_t *gpu_context = get_from_handle(&adapter->hdl_mgr, context);

    cm_wait_context_idle(gpu_context);

    return 0;
}

static int krnl_render(void *data, zx_render_t *render)
{
    int           ret;
    adapter_t     *adapter     = data;
    gpu_context_t *gpu_context = get_from_handle(&adapter->hdl_mgr, render->context);
    vidsch_render_t render_data = {0};

    render_data.command_length           = render->command_length;
    render_data.flags                    = render->flags;
    render_data.allocation_list_size     = render->allocation_count;
    render_data.patch_location_list_size = render->patch_location_count;
    render_data.sync_object_list_size    = render->sync_object_count;
    render_data.allocation_list          = (void*)(unsigned long)render->allocation_list;
    render_data.patch_location_list      = (void*)(unsigned long)render->patch_location_list;
    render_data.sync_object_list         = (void*)(unsigned long)render->sync_object_list;
    render_data.cmdbuf_count             = render->cmdbuf_count;
    render_data.render_counter           = gpu_context->render_counter++;
    render_data.cmdbuf_array             = (void*)(unsigned long)render->cmdbuf_array;

    ret = vidsch_render(gpu_context, &render_data);

    return ret;
}

static int krnl_create_fence_sync_object(void *data, zx_create_fence_sync_object_t *create, int binding)
{
    adapter_t     *adapter = data;
    gpu_device_t  *device  = get_from_handle(&adapter->hdl_mgr, create->device);
    unsigned int  status   = E_FAIL;

    status = vidsch_create_sync_object(device, create, binding);

    return status;
}

static int krnl_destroy_fence_sync_object(void *data, zx_destroy_fence_sync_object_t *destroy)
{
    adapter_t     *adapter = data;
    gpu_device_t  *device  = get_from_handle(&adapter->hdl_mgr, destroy->device);
    unsigned int  status   = S_OK;

    status = vidsch_destroy_sync_object(adapter, device, destroy->fence_sync_object);

    return status;
}

static int krnl_wait_fence_sync_object(void *data, zx_wait_fence_sync_object_t *wait)
{
    adapter_t     *adapter = data;
    gpu_context_t *context = get_from_handle(&adapter->hdl_mgr, wait->context);
    unsigned int  status   = S_OK;

    status = vidsch_wait_sync_object(context, wait);

    return status;
}

static int krnl_is_fence_object_signaled(void *data, unsigned int fence_sync_object, unsigned long long wait_value)
{
    adapter_t     *adapter = data;
    return vidsch_is_fence_signaled(adapter,fence_sync_object,wait_value);
}

static int krnl_is_fence_back(void *data, unsigned char engine_index, unsigned long long fence_id)
{
    adapter_t     *adapter = data;

    return vidsch_is_fence_back(adapter, engine_index, fence_id);
}

static int krnl_fence_value(void *data, zx_fence_value_t *value)
{
    adapter_t     *adapter = data;
    gpu_device_t  *device  = get_from_handle(&adapter->hdl_mgr, value->device);
    unsigned int  status   = E_FAIL;

    status = vidsch_fence_value(device, value);

    return status;
}

static int krnl_get_allocation_state(void *data, zx_get_allocation_state_t *state)
{
    adapter_t          *adapter      = data;
    vidmm_allocation_t *allocation   = get_from_handle(&adapter->hdl_mgr, state->hAllocation);
    unsigned int engine_mask = state->engine_mask;

    if (!allocation)
        return E_INVALIDARG;

    if (!vidsch_is_allocation_idle(adapter, engine_mask, allocation))
    {
        state->state = 1;
    }
    else
    {
        state->state = 0;
    }

    return 0;
}


static int krnl_add_hw_ctx_buf(void *data, zx_add_hw_ctx_buf_t *add)
{
    adapter_t        *adapter = data;
    gpu_device_t     *device  = get_from_handle(&adapter->hdl_mgr, add->device);
    gpu_context_t    *context = get_from_handle(&adapter->hdl_mgr, add->context);
    int ret = 0;

    ret = cm_add_hwctx_buffer(device, context, add->hw_ctx_buf_index);

    return ret;
}

static int  krnl_rm_hw_ctx_buf(void *data, zx_rm_hw_ctx_buf_t *rm)
{
    adapter_t        *adapter = data;
    gpu_device_t     *device  = get_from_handle(&adapter->hdl_mgr, rm->device);
    gpu_context_t    *context = get_from_handle(&adapter->hdl_mgr, rm->context);
    int ret = 0;

    ret = cm_remove_hwctx_buffer(device, context, rm->hw_ctx_buf_index);

    return ret;
}

static struct os_pages_memory* krnl_get_allocation_pages(void *data, int handle)
{
    adapter_t        *adapter = data;

    vidmm_allocation_t *allocation = get_from_handle(&adapter->hdl_mgr, handle);

    return allocation->pages_mem;
}

static int krnl_notify_interrupt(void *data, unsigned int interrupt_event)
{
    adapter_t        *adapter = data;

    return vidsch_notify_interrupt(adapter, interrupt_event);
}

static void krnl_prepare_and_mark_unpagable(void *data, unsigned int handle, zx_open_allocation_t *info)
{
    adapter_t           *adapter = data;
    vidmm_allocation_t  *allocation = get_from_handle(&adapter->hdl_mgr, handle);

    vidmm_prepare_and_mark_unpagable(adapter, allocation, info);
}

int krnl_query_dvfs_clamp(void *data, zx_dvfs_clamp_status_t* dvfs_clamp)
{
	adapter_t	  *adapter = data;
	int 		  status   = 0;

	status = vidsch_query_dvfs_clamp(adapter, dvfs_clamp);

	return status;
}

static int krnl_dvfs_set(void *data, zx_dvfs_set_t *dvfs_set)
{
    adapter_t     *adapter = data;
    int           status   = 0;

    status = vidsch_dvfs_set(adapter, dvfs_set);

    return status;
}

static void krnl_mark_pagable(void *data, unsigned int handle)
{
    adapter_t           *adapter = data;
    vidmm_allocation_t  *allocation = get_from_handle(&adapter->hdl_mgr, handle);

    vidmm_mark_pagable(adapter, allocation);
}

static void krnl_perf_event_add_isr_event(void *data, zx_perf_event_t *perf_event)
{
    adapter_t       *adapter = data;
    if (adapter->ctl_flags.perf_event_enable && adapter->perf_event_mgr)
    {
        perf_event_add_isr_event(adapter, perf_event);
    }
}

static void krnl_perf_event_add_event(void *data, zx_perf_event_t *perf_event)
{
    adapter_t       *adapter = data;
    if (adapter->ctl_flags.perf_event_enable && adapter->perf_event_mgr)
    {
        perf_event_add_event(adapter, perf_event);
    }
}

static void krnl_update_device_name(void *data, unsigned int device)
{
    adapter_t     *adapter = data;
    gpu_device_t  *gpu_device  = get_from_handle(&adapter->hdl_mgr, device);

    if (gpu_device)
    {
        gpu_device->pid = zx_get_current_pid();
        gpu_device->tid = zx_get_current_tid();
        util_get_process_name(gpu_device->pid, gpu_device->pname, ZX_MAX_PNAME_LEN);
    }
}

static void krnl_dump_hang_info_flag(void *data, int flag)
{
    adapter_t     *adapter = data;

    adapter->display_debugbus_flag = flag;
    zx_info("adapter->display_debugbus_flag = %d\n", flag);
}

static void krnl_set_timeout(void *data, unsigned int timeout)
{
    adapter_t *adapter = data;
    adapter->timeout_ms = timeout;
}
static core_interface_t zxe3k_gpu_core = {
#define INTERFACE(item) .item = krnl_##item
    INTERFACE(pre_init_adapter),
    INTERFACE(init_adapter),
    INTERFACE(deinit_adapter),
    INTERFACE(get_adapter_info),
    INTERFACE(update_adapter_info),
    INTERFACE(dump_resource),
    INTERFACE(debugfs_dump),
    INTERFACE(final_cleanup),
    INTERFACE(dvfs_set),
    INTERFACE(query_dvfs_clamp),
    INTERFACE(wait_chip_idle),
    INTERFACE(wait_allocation_idle),
    INTERFACE(get_allocation_state),
    INTERFACE(save_state),
    INTERFACE(restore_state),
    INTERFACE(query_info),
    INTERFACE(create_device),
    INTERFACE(destroy_device),
    INTERFACE(create_allocation),
    INTERFACE(create_allocation_list),
    INTERFACE(wait_context_idle),
    INTERFACE(create_allocation_from_pages),
    INTERFACE(destroy_allocation),
    INTERFACE(set_callback_func),
    INTERFACE(begin_perf_event),
    INTERFACE(end_perf_event),
    INTERFACE(get_perf_event),
    INTERFACE(send_perf_event),
    INTERFACE(get_perf_status),
    INTERFACE(begin_miu_dump_perf_event),
    INTERFACE(end_miu_dump_perf_event),
    INTERFACE(set_miu_reg_list_perf_event),
    INTERFACE(get_miu_dump_perf_event),
    INTERFACE(direct_get_miu_dump_perf_event),
    INTERFACE(create_context),
    INTERFACE(destroy_context),
    INTERFACE(render),
    INTERFACE(create_fence_sync_object),
    INTERFACE(destroy_fence_sync_object),
    INTERFACE(wait_fence_sync_object),
    INTERFACE(fence_value),
    INTERFACE(is_fence_object_signaled),
    INTERFACE(is_fence_back),
    INTERFACE(add_hw_ctx_buf),
    INTERFACE(rm_hw_ctx_buf),
    INTERFACE(cil2_misc),
    INTERFACE(get_allocation_pages),
    INTERFACE(notify_interrupt),
    INTERFACE(get_map_allocation_info),
    INTERFACE(prepare_and_mark_unpagable),
    INTERFACE(mark_pagable),
    INTERFACE(perf_event_add_isr_event),
    INTERFACE(perf_event_add_event),
    INTERFACE(update_device_name),
    INTERFACE(dump_hang_info_flag),
    INTERFACE(set_timeout),
#undef INTERFACE
};

core_interface_t *zx_core_interface = &zxe3k_gpu_core;
