/*************************************************************************************
* Copyright 1998-2014 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2014 S3 Graphics, Inc. All Rights Reserved.
* Copyright 2013-2014 VIA Alliance(Zhaoxin)Semiconductor Co., Ltd. All Rights Reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a
* copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sub license,
* and/or sell copies of the Software, and to permit persons to whom the
* Software is furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice (including the
* next paragraph) shall be included in all copies or substantial portions
* of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
* THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
* OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
* ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
* DEALINGS IN THE SOFTWARE.
*************************************************************************************/
#ifndef __ZX_TRACE_H__
#define __ZX_TRACE_H__

#ifdef ZX_USERMODE_TRACE

#include <inttypes.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/cdefs.h>
#include <sys/types.h>
#include <unistd.h>

#define CC_UNLIKELY( exp )  (__builtin_expect( !!(exp), false ))

/*
 default use c++11 <atomic> or C11 <stdatomic.h> in this header,
 but if the compiler is c++ without -std=c++11, which has no <atomic>,
 and g++ is incompatible with <stdatomic.h>, we should not use atomic
 in this header, or compile errors will happen.
 in this case, use extern zx_trace_init() instead of inline.
*/
#ifndef USE_ATOMIC_IN_TRACE_H
#define USE_ATOMIC_IN_TRACE_H 1
#endif

#ifdef __cplusplus
#if __cplusplus < 201103L
#undef USE_ATOMIC_IN_TRACE_H
#define USE_ATOMIC_IN_TRACE_H 0
#else /* __cplusplus < 201103L */
#include <atomic>
using std::atomic_bool;
using std::memory_order_acquire;
#endif
#else /* __cplusplus */
#include <stdatomic.h>
#endif

__BEGIN_DECLS

/**
 * The ZX_TRACE_TAG macro can be defined before including this header to trace
 * using one of the tags defined below.  It must be defined to one of the
 * following ZX_TRACE_TAG_* macros.  The trace tag is used to filter tracing in
 * userland to avoid some of the runtime cost of tracing when it is not desired.
 *
 * Defining ZX_TRACE_TAG to be ZX_TRACE_TAG_ALWAYS will result in the tracing always
 * being enabled - this should ONLY be done for debug code, as userland tracing
 * has a performance cost even when the trace is not being recorded.  Defining
 * ZX_TRACE_TAG to be ZX_TRACE_TAG_NEVER or leaving ZX_TRACE_TAG undefined will result
 * in the tracing always being disabled.
 *
 * Keep these in sync with catapult/systrace/systrace/tracing_agents/ftrace_agent.py.
 */
#define ZX_TRACE_TAG_NEVER            0       // This tag is never enabled.
#define ZX_TRACE_TAG_ALWAYS           (1<<0)  // This tag is always enabled.
#define ZX_TRACE_TAG_KEINTERFACE      (1<<1)
#define ZX_TRACE_TAG_XSERVER_DRIVER   (1<<2)
#define ZX_TRACE_TAG_CIL2_SERVICE     (1<<3)
#define ZX_TRACE_TAG_DP_SERVER        (1<<4)
#define ZX_TRACE_TAG_DP_BRIDGES       (1<<5)
#define ZX_TRACE_TAG_GLCORE           (1<<6)
#define ZX_TRACE_TAG_CLCORE           (1<<7)
#define ZX_TRACE_TAG_EGL              (1<<8)
#define ZX_TRACE_TAG_VPM              (1<<9)
#define ZX_TRACE_TAG_VA               (1<<10)
#define ZX_TRACE_TAG_VDPAU            (1<<11)
#define ZX_TRACE_TAG_LAST             ZX_TRACE_TAG_VDPAU

// Reserved for initialization.
#define ZX_TRACE_TAG_NOT_READY        (1ULL<<63)

#define ZX_TRACE_TAG_VALID_MASK ((ZX_TRACE_TAG_LAST - 1) | ZX_TRACE_TAG_LAST)

#ifndef ZX_TRACE_TAG
#define ZX_TRACE_TAG ZX_TRACE_TAG_NEVER
#elif ZX_TRACE_TAG > ZX_TRACE_TAG_VALID_MASK
#error ZX_TRACE_TAG must be defined to be one of the tags defined in cutils/trace.h
#endif

/**
 * Opens the trace file for writing and reads the property for initial tags.
 * The /sys/kernel/debug/dri/0/zx/umd_trace sets the tags to trace.
 * This function should not be explicitly called, the first call to any normal
 * trace function will cause it to be run safely.
 */
void zx_trace_setup();

/**
 * If tracing is ready, set zx_trace_enabled_tags to the debugfs file:
 * /sys/kernel/debug/dri/0/zx/umd_trace. Can be used as a tags change callback.
 */
void zx_trace_update_tags();

/**
 * Flag indicating whether setup has been completed, initialized to 0.
 * Nonzero indicates setup has completed.
 * Note: This does NOT indicate whether or not setup was successful.
 */
#if USE_ATOMIC_IN_TRACE_H
extern atomic_bool zx_trace_is_ready;
#endif

/**
 * Set of ZX_TRACE_TAG flags to trace for, initialized to ZX_TRACE_TAG_NOT_READY.
 * A value of zero indicates setup has failed.
 * Any other nonzero value indicates setup has succeeded, and tracing is on.
 */
extern uint64_t zx_trace_enabled_tags;

/**
 * Handle to the kernel's trace buffer, initialized to -1.
 * Any other value indicates setup has succeeded, and is a valid fd for tracing.
 */
extern int zx_trace_marker_fd;

/**
 * zx_trace_init readies the process for tracing by opening the trace_marker file.
 * Calling any trace function causes this to be run, so calling it is optional.
 * This can be explicitly run to avoid setup delay on first trace function.
 */
#define ZX_TRACE_INIT() zx_trace_init()
#if USE_ATOMIC_IN_TRACE_H
static inline void zx_trace_init()
{
    if (CC_UNLIKELY(!atomic_load_explicit(&zx_trace_is_ready, memory_order_acquire))) {
        zx_trace_setup();
    }
}
#else
extern void zx_trace_init();
#endif

/**
 * Get the mask of all tags currently enabled.
 * It can be used as a guard condition around more expensive trace calculations.
 * Every trace function calls this, which ensures zx_trace_init is run.
 */
#define ZX_TRACE_GET_ENABLED_TAGS() zx_trace_get_enabled_tags()
static inline uint64_t zx_trace_get_enabled_tags()
{
    zx_trace_init();
    return zx_trace_enabled_tags;
}

/**
 * Test if a given tag is currently enabled.
 * Returns nonzero if the tag is enabled, otherwise zero.
 * It can be used as a guard condition around more expensive trace calculations.
 */
#define ZX_TRACE_ENABLED() zx_trace_is_tag_enabled(ZX_TRACE_TAG)
static inline uint64_t zx_trace_is_tag_enabled(uint64_t tag)
{
    return zx_trace_get_enabled_tags() & tag;
}

/**
 * Trace the beginning of a context.  name is used to identify the context.
 * Optional args can be appended after a '|', args will be added to the slice.
 * This is often used to time function execution.
 */
#define ZX_TRACE_BEGIN(name, ...) zx_trace_begin(ZX_TRACE_TAG, name, ##__VA_ARGS__)
static inline void zx_trace_begin(uint64_t tag, const char* name, ...)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_begin_body(const char*, va_list);
        va_list args;
        va_start(args, name);
        zx_trace_begin_body(name, args);
        va_end(args);
    }
}

/**
 * Trace additional args of a context.
 * Some args are only known in the middle or end of the execution.
 * This should match up (and occur after) a corresponding ZX_TRACE_BEGIN.
 */
#define ZX_TRACE_ARGS(format, ...) zx_trace_args(ZX_TRACE_TAG, format, ##__VA_ARGS__)
static inline void zx_trace_args(uint64_t tag, const char* format, ...)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_args_body(const char*, va_list);
        va_list args;
        va_start(args, format);
        zx_trace_args_body(format, args);
        va_end(args);
    }
}

/**
 * Trace the end of a context.
 * This should match up (and occur after) a corresponding ZX_TRACE_BEGIN.
 */
#define ZX_TRACE_END() zx_trace_end(ZX_TRACE_TAG)
static inline void zx_trace_end(uint64_t tag)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_end_body();
        zx_trace_end_body();
    }
}

/**
 * Trace the end of a context, with extra args.
 * Some args are only known in the middle or end of the execution.
 * This should match up (and occur after) a corresponding ZX_TRACE_BEGIN.
 */
#define ZX_TRACE_END2(format, ...) zx_trace_end2(ZX_TRACE_TAG, format, ##__VA_ARGS__)
static inline void zx_trace_end2(uint64_t tag, const char* format, ...)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_end2_body(const char*, va_list);
        va_list args;
        va_start(args, format);
        zx_trace_end2_body(format, args);
        va_end(args);
    }
}

/**
 * Trace the beginning of an asynchronous event. Unlike ZX_TRACE_BEGIN/ZX_TRACE_END
 * contexts, asynchronous events do not need to be nested. The name describes
 * the event, and the cookie provides a unique identifier for distinguishing
 * simultaneous events. The name and cookie used to begin an event must be
 * used to end it.
 */
#define ZX_TRACE_ASYNC_BEGIN(name, cookie) \
    zx_trace_async_begin(ZX_TRACE_TAG, name, cookie)
static inline void zx_trace_async_begin(uint64_t tag, const char* name,
        int32_t cookie)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_async_begin_body(const char*, int32_t);
        zx_trace_async_begin_body(name, cookie);
    }
}

/**
 * Trace the end of an asynchronous event.
 * This should have a corresponding ZX_TRACE_ASYNC_BEGIN.
 */
#define ZX_TRACE_ASYNC_END(name, cookie) zx_trace_async_end(ZX_TRACE_TAG, name, cookie)
static inline void zx_trace_async_end(uint64_t tag, const char* name, int32_t cookie)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_async_end_body(const char*, int32_t);
        zx_trace_async_end_body(name, cookie);
    }
}

/**
 * Traces an integer counter value.  name is used to identify the counter.
 * This can be used to track how a value changes over time.
 */
#define ZX_TRACE_INT(name, value) zx_trace_int(ZX_TRACE_TAG, name, value)
static inline void zx_trace_int(uint64_t tag, const char* name, int32_t value)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_int_body(const char*, int32_t);
        zx_trace_int_body(name, value);
    }
}

/**
 * Traces a 64-bit integer counter value.  name is used to identify the
 * counter. This can be used to track how a value changes over time.
 */
#define ZX_TRACE_INT64(name, value) zx_trace_int64(ZX_TRACE_TAG, name, value)
static inline void zx_trace_int64(uint64_t tag, const char* name, int64_t value)
{
    if (CC_UNLIKELY(zx_trace_is_tag_enabled(tag))) {
        void zx_trace_int64_body(const char*, int64_t);
        zx_trace_int64_body(name, value);
    }
}

__END_DECLS

// enable ZX_TRACE_NAME and ZX_TRACE_CALL for C++
#ifdef __cplusplus

class ScopedTrace {
public:
inline ScopedTrace(uint64_t tag, const char* name)
    : mTag(tag) {
    zx_trace_begin(mTag,name);
}

inline ~ScopedTrace() {
    zx_trace_end(mTag);
}

private:
    uint64_t mTag;
};

// ZX_TRACE_NAME traces the beginning and end of the current scope.  To trace
// the correct start and end times this macro should be declared first in the
// scope body.
#define ZX_TRACE_NAME(name) ScopedTrace ___tracer(ZX_TRACE_TAG, name)
// ZX_TRACE_CALL is an ZX_TRACE_NAME that uses the current function name.
#define ZX_TRACE_CALL() ZX_TRACE_NAME(__FUNCTION__)

#else // __cplusplus
#define ZX_TRACE_NAME(...)
#define ZX_TRACE_CALL()

#endif // __cplusplus

#else // ZX_USERMODE_TRACE

#define ZX_TRACE_INIT()
#define ZX_TRACE_GET_ENABLED_TAGS()
#define ZX_TRACE_ENABLED() 0
#define ZX_TRACE_BEGIN(name, ...)
#define ZX_TRACE_ARGS(format, ...)
#define ZX_TRACE_END()
#define ZX_TRACE_END2(format, ...)
#define ZX_TRACE_ASYNC_BEGIN(name, cookie)
#define ZX_TRACE_ASYNC_END(name, cookie)
#define ZX_TRACE_INT(name, value)
#define ZX_TRACE_NAME(...)
#define ZX_TRACE_CALL()

#endif // ZX_USERMODE_TRACE

#endif // __ZX_TRACE_H__
