/*
 * Copyright (c) 2018 Shanghai Zhaoxin Semiconductor Co., Ltd.
 * All Rights Reserved.
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Shanghai Zhaoxin Semiconductor Co., Ltd.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Shanghai Zhaoxin Semiconductor Co., Ltd.
 *
 * The copyright of the source code is protected by the copyright laws of the People's
 * Republic of China and the related laws promulgated by the People's Republic of China
 * and the international covenant(s) ratified by the People's Republic of China.
 *
 */
#include "zx_atomic.h"

#if DRM_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)

struct drm_atomic_state* zx_atomic_state_alloc(struct drm_device *dev)
{
	zx_ato_state_t *state = zx_calloc(sizeof(zx_ato_state_t));

	if (!state || drm_atomic_state_init(dev, &state->base_ato_state) < 0) 
        {
		zx_free(state);
		return NULL;
	}

	return &state->base_ato_state;
}

void zx_atomic_state_clear(struct drm_atomic_state *s)
{
	zx_ato_state_t *state = to_zx_atomic_state(s);
	drm_atomic_state_default_clear(s);
}

void zx_atomic_state_free(struct drm_atomic_state *s)
{
    	zx_ato_state_t *state = to_zx_atomic_state(s);
	drm_atomic_state_default_release(s);
        zx_free(state);
}

struct drm_crtc_state* zx_crtc_duplicate_state(struct drm_crtc *crtc)
{
    zx_crtc_state_t* crtc_state;

    crtc_state = zx_calloc(sizeof(zx_crtc_state_t));
    if (!crtc_state)
    {
        return NULL;
    }

    __drm_atomic_helper_crtc_duplicate_state(crtc, &crtc_state->base_cstate);

    return &crtc_state->base_cstate;
}

void zx_crtc_destroy_state(struct drm_crtc *crtc, struct drm_crtc_state *s)
{
    zx_crtc_state_t*  state = to_zx_crtc_state(s);
    __drm_atomic_helper_crtc_destroy_state(s);
    zx_free(state);
}

#if DRM_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
int zx_crtc_helper_check(struct drm_crtc *crtc, struct drm_atomic_state *crtc_state)
#else
int zx_crtc_helper_check(struct drm_crtc *crtc, struct drm_crtc_state *state)
#endif

{
    //a patch for HDMI & CRT on IGA1, when switch from IGA1->HDMI1 to IGA1->CRT with same timing, 
    //framework will not call crtc/encoder_set_mode, so we will meet active device/path error

#if DRM_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
    struct drm_crtc_state *state = drm_atomic_get_old_crtc_state(crtc_state, crtc);
#endif

    if(!state || state->mode_changed || !state->active)
    {
        return 0;
    }

    if(!drm_mode_equal(&state->adjusted_mode, &crtc->state->adjusted_mode))
    {
        zx_info("Adj mode changed on crtc-%d, need do full modeset.\n", crtc->index);
        state->mode_changed = 1;
    }
    else if(state->connectors_changed)
    {
        zx_info("Connectors changed on crtc-%d, need do full modeset.\n", crtc->index);
        state->mode_changed = 1;
    }    

    return 0;
}

struct drm_connector_state* zx_connector_duplicate_state(struct drm_connector *connector)
{
    zx_connector_state_t* zx_conn_state;

    zx_conn_state = zx_calloc(sizeof(zx_connector_state_t));
    if (!zx_conn_state)
    {
        return NULL;
    }

    __drm_atomic_helper_connector_duplicate_state(connector, &zx_conn_state->base_conn_state);

    return &zx_conn_state->base_conn_state;
}

void zx_connector_destroy_state(struct drm_connector *connector, struct drm_connector_state *state)
{
    zx_connector_state_t *zx_conn_state = to_zx_conn_state(state);
    __drm_atomic_helper_connector_destroy_state(state);
    zx_free(zx_conn_state);
}

void zx_atomic_helper_commit_tail(struct drm_atomic_state *old_state)
{
    struct drm_device *dev = old_state->dev;
    zx_card_t*  zx_card = dev->dev_private;
    disp_info_t*  disp_info = (disp_info_t*)zx_card->disp_info;
    psr_data_t* psr_data = disp_info->psr_data;

#if DRM_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
    uint32_t flags = DRM_PLANE_COMMIT_NO_DISABLE_AFTER_MODESET;
#else
    bool flags = false;
#endif

#if ENABLE_RUNTIME_PM
   int ret;
   bool active = false;
   struct drm_crtc *crtc;
    ret = pm_runtime_get_sync(dev->dev);
    if (ret < 0) {
        pm_runtime_mark_last_busy(dev->dev);
        pm_runtime_put_autosuspend(dev->dev);
        return ;
    }
#endif

    psr_acquire_display(psr_data, PSR_FLIP_REF);

    drm_atomic_helper_commit_modeset_disables(dev, old_state);

    drm_atomic_helper_commit_modeset_enables(dev, old_state);

    drm_atomic_helper_commit_planes(dev, old_state, flags);

    drm_atomic_helper_commit_hw_done(old_state);

#if DRM_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
    drm_atomic_helper_fake_vblank(old_state);
#endif

#if DRM_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
    drm_atomic_helper_wait_for_flip_done(dev, old_state);
#else
    drm_atomic_helper_wait_for_vblanks(dev, old_state);
#endif

    drm_atomic_helper_cleanup_planes(dev, old_state);

    psr_release_display(psr_data, PSR_FLIP_REF);

#if ENABLE_RUNTIME_PM
    list_for_each_entry(crtc, &dev->mode_config.crtc_list, head) {
        if (crtc->enabled) {
            active = true;
        }
    }
    pm_runtime_mark_last_busy(dev->dev);
    /* if we have active crtcs and we don't have a power ref,take the current one */
    if (active && !zx_card->have_disp_power_ref) {
        zx_info("Runtime PM: have active crtcs and reference drm device\n");
        zx_card->have_disp_power_ref = true;
        return ;
    }
    /* if we have no active crtcs, then drop the power ref  we got before */
    if(!active && zx_card->have_disp_power_ref) {
        zx_info("Runtime PM: have no active crtcs ,drop the power ref we got before\n");
        pm_runtime_put_autosuspend(dev->dev);
        zx_card->have_disp_power_ref = false;
    }
    pm_runtime_put_autosuspend(dev->dev);
#endif

}
#endif

