/*
 * Copyright (c) 2018 Shanghai Zhaoxin Semiconductor Co., Ltd.
 * All Rights Reserved.
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Shanghai Zhaoxin Semiconductor Co., Ltd.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Shanghai Zhaoxin Semiconductor Co., Ltd.
 *
 * The copyright of the source code is protected by the copyright laws of the People's
 * Republic of China and the related laws promulgated by the People's Republic of China
 * and the international covenant(s) ratified by the People's Republic of China.
 *
 */
#include "os_interface.h"
#include "zx_driver.h"
#include "zx_ioctl.h"
#include "zx_debugfs.h"
#include "zx_fence.h"
#include "zx_irq.h"
#include "zx_fbdev.h"

extern int zx_run_on_qt;

static int zx_init_procfs_entry(zx_card_t *zx);
static int zx_deinit_procfs_entry(zx_card_t *zx);

pgprot_t os_get_pgprot_val(unsigned int *cache_type, pgprot_t old_prot, int io_map)
{
    pgprot_t prot = old_prot;

    if(*cache_type == ZX_MEM_UNCACHED)
    {   
        prot = pgprot_noncached(old_prot);
    }
    else if(*cache_type == ZX_MEM_WRITE_COMBINED)
    {
#ifdef CONFIG_X86
#ifdef CONFIG_X86_PAT
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,19,0)
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | cachemode2protval(_PAGE_CACHE_MODE_WC));
#else
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | _PAGE_CACHE_WC);
#endif
#else
        prot = pgprot_noncached(old_prot);

        *cache_type = ZX_MEM_UNCACHED;	
#endif
#else
        prot = pgprot_writecombine(old_prot);
#endif
        
    }
#ifdef ZX_AHB_BUS
    else if(*cache_type == ZX_MEM_WRITE_BACK)
    {   
        // for snoop path, current linux kennel defalut is not WA+WB
        prot = __pgprot((pgprot_val(old_prot)&(~(L_PTE_MT_MASK))) | (L_PTE_MT_WRITEBACK | L_PTE_MT_WRITEALLOC));
    }
#endif
    return prot;
}

#ifndef __frv__
int zx_map_system_io(struct vm_area_struct* vma, zx_map_argu_t *map)
{
    unsigned int  cache_type = map->flags.cache_type;
#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 1);

    map->flags.cache_type = cache_type;

    if (remap_pfn_range(vma, vma->vm_start,
                vma->vm_pgoff,
                vma->vm_end - vma->vm_start,
                vma->vm_page_prot))
    {
        return -EAGAIN;
    }

    return 0;    
}
#if 1
int zx_map_system_ram(struct vm_area_struct* vma, zx_map_argu_t *map)
{
    struct os_pages_memory *memory = map->memory;	
    unsigned long start = vma->vm_start;
    unsigned long pfn;
    unsigned int  cache_type, curr_block_offset;
	
    int i;
    int start_page  = _ALIGN_DOWN(map->offset, PAGE_SIZE) / PAGE_SIZE;
    int end_page    = start_page + PAGE_ALIGN(map->size) / PAGE_SIZE;

    
#if LINUX_VERSION_CODE <= 0x02040e
    vma->vm_flags |= VM_LOCKED;
#elif LINUX_VERSION_CODE < 0x030700
    vma->vm_flags |= VM_RESERVED;
#else
    vma->vm_flags |= VM_DONTEXPAND | VM_DONTDUMP;
#endif

    cache_type = map->flags.cache_type;

#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 0);
    for (i = start_page; i < end_page; i++)
    {
        struct page *pg = memory->pages[i];
        pfn = page_to_pfn(pg);
        if(remap_pfn_range(vma, start, pfn, PAGE_SIZE, vma->vm_page_prot))
        {
            return -EAGAIN;
        }
        start += PAGE_SIZE;
    }

    return 0;
}
#endif
#endif

static int zx_notify_fence_interrupt(zx_card_t *zx)
{
    return zx_core_interface->notify_interrupt(zx->adapter, INT_FENCE);
}

void zx_interrupt_init(zx_card_t *zx)
{
    tasklet_init(&zx->fence_notify, (void*)zx_notify_fence_interrupt, (unsigned long)zx);
    tasklet_disable(&zx->fence_notify);
    tasklet_enable(&zx->fence_notify);
}

void zx_interrupt_reinit(zx_card_t *zx)
{
    tasklet_enable(&zx->fence_notify); 
}

void zx_interrupt_deinit(zx_card_t *zx)
{
    tasklet_disable(&zx->fence_notify);
    synchronize_irq(to_pci_dev(zx->drm_dev->dev)->irq);
    tasklet_kill(&zx->fence_notify);
}

void zx_enable_interrupt(void *pdev_)
{
    struct drm_device *dev = _to_drm_device(pdev_);
    zx_card_t *zx = dev->dev_private;

    zx_disp_enable_interrupt(zx->disp_info);
}

void zx_disable_interrupt(void *pdev_)
{
    struct drm_device *dev = _to_drm_device(pdev_);
    zx_card_t *zx = dev->dev_private;

    zx_disp_disable_interrupt(zx->disp_info);
}

static krnl_import_func_list_t zx_export =
{
    .udelay                        = zx_udelay,
    .begin_timer                   = zx_begin_timer,
    .end_timer                     = zx_end_timer,
    .do_div                        = zx_do_div,
    .msleep                        = zx_msleep,
    .getsecs                       = zx_getsecs,
    .get_nsecs                     = zx_get_nsecs,
    .assert                        = zx_assert,
    .dump_stack                    = zx_dump_stack,
    .copy_from_user                = zx_copy_from_user,
    .copy_to_user                  = zx_copy_to_user,
    .slowbcopy_tobus               = zx_slowbcopy_tobus,
    .slowbcopy_frombus             = zx_slowbcopy_frombus,
    .memset                        = zx_memset,
    .memcpy                        = zx_memcpy,
    .memcmp_priv                   = zx_memcmp,
    .byte_copy                     = zx_byte_copy,
    .strcmp                        = zx_strcmp,
    .strcpy                        = zx_strcpy,
    .strncmp                       = zx_strncmp,
    .strncpy                       = zx_strncpy,
    .strlen                        = zx_strlen,
    .read64                        = zx_read64,
    .read32                        = zx_read32,
    .read16                        = zx_read16,
    .read8                         = zx_read8,
    .write32                       = zx_write32,
    .write16                       = zx_write16,
    .write8                        = zx_write8,
    .memsetio                      = zx_memsetio,
    .memcpy_fromio                 = zx_memcpy_fromio,
    .memcpy_toio                   = zx_memcpy_toio,
    .file_open                     = zx_file_open,
    .file_close                    = zx_file_close,
    .file_read                     = zx_file_read,
    .file_write                    = zx_file_write,
    .vsprintf                      = zx_vsprintf,
    .vsnprintf                     = zx_vsnprintf,
    .sscanf                        = zx_sscanf,
    .printk                        = zx_printk,
    .cb_printk                     = zx_cb_printk,
    .seq_printf                    = zx_seq_printf,
    .find_first_zero_bit           = zx_find_first_zero_bit,
    .find_next_zero_bit            = zx_find_next_zero_bit,
    .set_bit                       = zx_set_bit,
    .clear_bit                     = zx_clear_bit,
    .getsecs                       = zx_getsecs,
    .get_nsecs                     = zx_get_nsecs,
    .create_thread                 = zx_create_thread,
    .destroy_thread                = zx_destroy_thread,
    .thread_should_stop            = zx_thread_should_stop,
    .create_atomic                 = zx_create_atomic,
    .destroy_atomic                = zx_destroy_atomic,
    .atomic_read                   = zx_atomic_read,
    .atomic_inc                    = zx_atomic_inc,
    .atomic_dec                    = zx_atomic_dec,
    .atomic_add                    = zx_atomic_add,
    .atomic_sub                    = zx_atomic_sub,
    .create_mutex                  = zx_create_mutex,
    .destroy_mutex                 = zx_destroy_mutex,
    .mutex_lock                    = zx_mutex_lock,
    .mutex_lock_killable           = zx_mutex_lock_killable,
    .mutex_trylock                 = zx_mutex_trylock,
    .mutex_unlock                  = zx_mutex_unlock,
    .create_sema                   = zx_create_sema,
    .destroy_sema                  = zx_destroy_sema,
    .down                          = zx_down,
    .down_trylock                  = zx_down_trylock,
    .up                            = zx_up,
    .create_rwsema                 = zx_create_rwsema,
    .destroy_rwsema                = zx_destroy_rwsema,
    .down_read                     = zx_down_read,
    .down_write                    = zx_down_write,
    .up_read                       = zx_up_read,
    .up_write                      = zx_up_write,
    .create_spinlock               = zx_create_spinlock,
    .destroy_spinlock              = zx_destroy_spinlock,
    .spin_lock                     = zx_spin_lock,
    .spin_unlock                   = zx_spin_unlock,
    .spin_lock_irqsave             = zx_spin_lock_irqsave,
    .spin_unlock_irqrestore        = zx_spin_unlock_irqrestore,
    .create_event                  = zx_create_event,
    .destroy_event                 = zx_destroy_event,
    .wait_event_thread_safe        = zx_wait_event_thread_safe,
    .wait_event                    = zx_wait_event,
    .wake_up_event                 = zx_wake_up_event,
    .thread_wait                   = zx_thread_wait,
    .create_wait_queue         = zx_create_wait_queue,
    .wait_for_events             = thread_wait_for_events,
    .thread_wake_up                = zx_thread_wake_up,
    ._thread_wake_up              = general_thread_wake_up,
    .dump_stack                    = zx_dump_stack,
    .try_to_freeze                 = zx_try_to_freeze,
    .freezable                     = zx_freezable,
    .clear_freezable               = zx_clear_freezable,
    .set_freezable                 = zx_set_freezable,
    .freezing                      = zx_freezing,
    .get_current_pid               = zx_get_current_pid,
    .get_current_tid               = zx_get_current_tid,
    .get_current_pname             = zx_get_current_pname,
    .flush_cache                   = zx_flush_cache,
    .inv_cache                     = zx_inv_cache,
    .pages_memory_for_each_continues = zx_pages_memory_for_each_continues,
    .ioremap                       = zx_ioremap,
    .iounmap_priv                  = zx_iounmap,
    .mtrr_add                      = zx_mtrr_add,
    .mtrr_del                      = zx_mtrr_del,
    .get_mem_info                  = zx_get_mem_info,
    .register_shrinker             = zx_register_shrinker,
    .unregister_shrinker           = zx_unregister_shrinker,
    .pages_memory_swapout          = zx_pages_memory_swapout,
    .pages_memory_swapin           = zx_pages_memory_swapin,
    .release_file_storage          = zx_release_file_storage,
    .get_bus_config                = zx_get_bus_config,
    .get_platform_config           = zx_get_platform_config,
    .get_command_status16          = zx_get_command_status16,
    .write_command_status16        = zx_write_command_status16,
    .get_command_status32          = zx_get_command_status32,
    .write_command_status32        = zx_write_command_status32,
    .get_rom_start_addr            = zx_get_rom_start_addr,
    .get_rom_save_addr             = zx_get_rom_save_addr,
    .write_rom_save_addr           = zx_write_rom_save_addr,
    .get_bar1                      = zx_get_bar1,
    .logmsg_init                   = zx_logmsg_init,
    .logmsg                        = zx_logmsg,
    .logmsg_deinit                 = zx_logmsg_deinit,

    .malloc_track                  = zx_malloc_track,
    .calloc_track                  = zx_calloc_track,
    .free_track                    = zx_free_track,
    .malloc_priv                   = zx_malloc_priv,
    .calloc_priv                   = zx_calloc_priv,
    .free_priv                     = zx_free_priv,
    .allocate_pages_memory_track   = zx_allocate_pages_memory_track,
    .free_pages_memory_track       = zx_free_pages_memory_track,
    .allocate_pages_memory_priv    = zx_allocate_pages_memory_priv,
    .free_pages_memory_priv        = zx_free_pages_memory_priv,

    .map_pages_memory_track        = zx_map_pages_memory_track,
    .unmap_pages_memory_track      = zx_unmap_pages_memory_track,
    .map_pages_memory_priv         = zx_map_pages_memory_priv,
    .unmap_pages_memory_priv       = zx_unmap_pages_memory_priv,
    .map_io_memory_track           = zx_map_io_memory_track,
    .unmap_io_memory_track         = zx_unmap_io_memory_track,
    .map_io_memory_priv            = zx_map_io_memory_priv,
    .unmap_io_memory_priv          = zx_unmap_io_memory_priv,
    .register_trace_events         = zx_register_trace_events,
    .unregister_trace_events       = zx_unregister_trace_events,
    .task_create_trace_event       = zx_task_create_trace_event,
    .task_submit_trace_event       = zx_task_submit_trace_event,
    .fence_back_trace_event        = zx_fence_back_trace_event,
    .begin_section_trace_event     = zx_begin_section_trace_event,
    .end_section_trace_event       = zx_end_section_trace_event,
    .counter_trace_event           = zx_counter_trace_event,

    .query_platform_caps           = zx_query_platform_caps,

    .console_lock                  = zx_console_lock,
    .enable_interrupt              = zx_enable_interrupt,
    .disable_interrupt             = zx_disable_interrupt,
    .os_printf                     = zx_printf,
    .disp_wait_idle                = zx_disp_wait_idle,
    .mb                            = zx_mb,
    .rmb                           = zx_rmb,
    .wmb                           = zx_wmb,
    .flush_wc                      = zx_flush_wc,
    .dsb                           = zx_dsb,

    .get_pages_from_userptr        = zx_get_pages_from_userptr,

    .set_cpus_allowed_ptr          = zx_set_cpus_allowed_ptr,
    
    .setup_timer                   = zx_setup_timer,
    .mod_timer                     = zx_mod_timer,
    .del_timer                     = zx_del_timer,
    .destroy_timer                 = zx_destroy_timer,
};

/***/
#include <linux/proc_fs.h>

#ifdef NO_PROC_CREATE_FUNC
static int zx_read_proc(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{

    zx_card_t    *zx = data;
    struct os_printer p = zx_info_printer(zx);
    zx_core_interface->dump_resource(&p, zx->adapter, 5, 0);

    return 0;
}
#else
static ssize_t zx_proc_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
    zx_card_t *zx = PDE_DATA(file_inode(filp));
    struct os_printer p = zx_info_printer(zx);
    zx_core_interface->dump_resource(&p, zx->adapter, 1, 0);

    return 0;
}

#define ZX_DUMP_FENCE_INDEX           0x5000
#define ZX_CLEAR_FENCE_INDEX          0x6000
#define ZX_FORCE_WAKEUP               0x7000
#define ZX_FORCE_QUIT                 0x8000

#define ZX_DUMP_COMPRESS_PS_INDEX    (0x100 + ZX_STREAM_PS)
#define ZX_DUMP_COMPRESS_SS_INDEX    (0x100 + ZX_STREAM_SS)
#define ZX_DUMP_COMPRESS_TS_INDEX    (0x100 + ZX_STREAM_TS)
#define ZX_DUMP_COMPRESS_4S_INDEX    (0x100 + ZX_STREAM_4S)
#define ZX_DUMP_COMPRESS_5S_INDEX    (0x100 + ZX_STREAM_5S)
#define ZX_DUMP_COMPRESS_6S_INDEX    (0x100 + ZX_STREAM_6S)
#define ZX_DUMP_COMPRESS_MAX_INDEX   (0x100 + ZX_STREAM_MAX)


static ssize_t zx_proc_write(struct file *filp, const char __user *buf, size_t count, loff_t *offp)
{
    zx_card_t *zx = PDE_DATA(file_inode(filp));

    char mode = '0';
    int dump_index = 0;
    unsigned int ret = 0;
    char temp[20] = {0};
    unsigned int interval = 0, crtc = 0;
    unsigned long value = 0;

//maybe later we can change it to using zx_seq_file_printer
    struct os_printer p = zx_info_printer(zx);

    if(count > 1 && count < 20)
    {
        ret = copy_from_user(&temp[0], buf, count);
        temp[count-1] = '\0';

        if(count == 2)
        {
            mode = temp[0];

            if(mode >= '0' && mode <= '9')
            {
                dump_index = (mode - '0');
            }
            else if((mode >= 'a' && mode < 'e'))
            {
                dump_index =  mode - 'a' + 0xa;
            }
            else if (mode == 'f')
            {
                dump_index = ZX_DUMP_FENCE_INDEX;
            }
            else if (mode == 'g')
            {
                dump_index = ZX_CLEAR_FENCE_INDEX;
            }
            else if (mode == 'w')
            {
                dump_index = ZX_FORCE_WAKEUP;
            }
            else if (mode == 'q')
            {
                dump_index = ZX_FORCE_QUIT;
            }
            else
            {
               zx_error("echo error proc mode, should between 0 and d \n");
               return -EFAULT;
            }
        }
        else
        {
            value = simple_strtoul(&temp[1], NULL, 16);

            zx_info("value %d \n", value);

            if((temp[0] == 'c' || temp[0] == 'C') &&
                    (temp[2] == 's' || temp[2] == 'S'))
            {
                if(zx_strlen(temp) != 4)
                {
                    zx_info("invalid flag buffer dump parameter!\n");
                }

                crtc = simple_strtoul(&temp[1], NULL, 16);
                dump_index = simple_strtoul(&temp[3], NULL, 16);

                dump_index += ZX_DUMP_COMPRESS_PS_INDEX;
            }
            else
            {
               zx_error("invalid index\n");
            }
         }
    }
    else
    {
        zx_error("proc write pass too long param, should less then 20\n");
    }

    if (dump_index == ZX_DUMP_FENCE_INDEX)
    {
        zx_dma_track_fences_dump(zx);
    }
    else if (dump_index == ZX_CLEAR_FENCE_INDEX)
    {
        zx_dma_track_fences_clear(zx);
    }
    else if((dump_index >=0 && dump_index <=5))
    {
        zx_core_interface->dump_resource(&p, zx->adapter, dump_index, 0);
    }
    else if (dump_index == ZX_FORCE_WAKEUP)
    {
        zx_core_interface->notify_interrupt(zx->adapter, INT_FENCE);
        zx_core_interface->dump_resource(&p, zx->adapter, dump_index, 0);
    }

    return count;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,6,0)
static const struct proc_ops zx_proc_fops =
{
    .proc_read  = zx_proc_read,
    .proc_write = zx_proc_write,
};

#else
static const struct file_operations zx_proc_fops =
{
    .owner = THIS_MODULE,
    .read  = zx_proc_read,
    .write = zx_proc_write,
};
#endif
#endif

void zx_card_pre_init(zx_card_t *zx, void *pdev)
{
    zx->pdev  = pdev;
    
    zx->zxfb_enable = 0;
#ifdef CONFIG_FB
    zx->zxfb_enable = zx_fb;
#endif

    if(zx->zxfb_enable)
    {
         zx_fbdev_disable_vesa(zx);
    }

    return;
}

int zx_card_init(zx_card_t *zx, void *pdev)
{
    unsigned short command = 0;
    int ret = 0, primary;

    zx_init_procfs_entry(zx);

    zx_init_bus_id(zx);

    zx_get_command_status16(pdev, &command);

    primary = (command & 0x01) ? 1 : 0;

    /* we can enable zxfb primay and no fb driver enable by default */
    if(!primary)
    {
        zx->zxfb_enable = FALSE;
    }

#ifdef ZX_PCIE_BUS
    zx->support_msi = 1;
#else
    zx->support_msi = 0;
#endif

#ifdef CONFIG_X86
#ifdef CONFIG_HYPERVISOR_GUEST
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,9)
    if(x86_hyper)
    {
        zx_info("detect hypervisor: %s\n", x86_hyper->name);

        zx->adapter_info.run_on_hypervisor = 1;
    }
#else
    if(!hypervisor_is_type(X86_HYPER_NATIVE))
    {
        zx_info("detect hypervisor: %d\n", x86_hyper_type);

        zx->adapter_info.run_on_hypervisor = 1;
    }
#endif
#endif
#endif

#if 0
    // keep head range memory unused for host use if driver run in virtual machine,
    // because host always use this head memory range as frame buffer, guest should
    // not touch this memory.
    if(zx->adapter_info.run_on_hypervisor)
    {
        // patch reserved video memory size.
        zx->reserved_vmem += vmem_reserve_for_host;
    }
#endif

    zx->adapter = zx_core_interface->pre_init_adapter(pdev, &zx->a_info, &zx_export);

    if(zx->adapter == NULL)
    {
        ret = -1;

        zx_error("init adapter failed\n");
    }

    // this should be put before init mode set, since init mode set will
    // set up interrrupt, and if dvfs interrupt comes, and dvfs_info not allocated
    // may cause kernel crash

    zx->dvfs_info = zx_calloc(sizeof(zx_dvfs_info_t));
    if(zx->dvfs_info == NULL)
    {
        ret = -1;

        zx_error("init adapter dvfs info failed\n");
    }

    zx_init_modeset(zx->drm_dev);

    zx_init_psr(zx->drm_dev);

    zx_core_interface->init_adapter(zx->adapter, zx->reserved_vmem, zx->disp_info);


#ifndef ZX_HW_NULL
    zx_interrupt_init(zx);
#endif

    zx->lock = zx_create_mutex();

    zx->debugfs_dev = zx_debugfs_create(zx, zx->drm_dev->primary->debugfs_root);

    return ret;
}

int zx_card_deinit(zx_card_t *zx)
{
    char *name;
    if(zx->zxfb_enable)
    {
        zx_fbdev_deinit(zx);
    }

    zx_core_interface->wait_chip_idle(zx->adapter);

    zx_deinit_psr(zx->drm_dev);

    zx_deinit_modeset(zx->drm_dev);

#ifndef ZX_HW_NULL
    zx_interrupt_deinit(zx);
#endif
    zx_core_interface->deinit_adapter(zx->adapter);

    zx->len = 0;

    if(zx->lock)
    {
        zx_destroy_mutex(zx->lock);
        zx->lock = NULL;
    }

    if(zx->dvfs_info)
    {
        zx_free(zx->dvfs_info);
    }

    if(zx->debugfs_dev)
    {
        zx_debugfs_destroy((zx_debugfs_device_t*)(zx->debugfs_dev));
        zx->debugfs_dev = NULL;
    }

    if(zx->procfs_entry)
    {
        zx_deinit_procfs_entry(zx);
        zx->procfs_entry = NULL;
    }

    return 0;
}

static int zx_init_procfs_entry(zx_card_t *zx)
{
    char name[64] = "";

    zx_vsnprintf(name, 64, "driver/dri%d", zx->index);

#ifdef NO_PROC_CREATE_FUNC
    //create_proce_read_entry is not used any more in Linux 3.10 and above. use proc_create_data instead.
    zx->procfs_entry = create_proc_read_entry(name, 0, NULL, zx_read_proc, zx);
#else
    zx->procfs_entry = proc_create_data(name, 0, NULL, &zx_proc_fops, zx);
#endif

    zx_info("%s() name-%s, procfs_entry-0x%x\n", ZX_FUNC_NAME(__func__), name, zx->procfs_entry);

    return 0;
}

static int zx_deinit_procfs_entry(zx_card_t *zx)
{
    char name[64] = "";

    zx_vsnprintf(name, 64, "driver/dri%d", zx->index);
#ifdef NO_PROC_CREATE_FUNC
    //this path is not verified yet
    remove_proc_entry(name, zx->procfs_entry);
#else
    proc_remove(zx->procfs_entry);
#endif

    return 0;
}

