/*
 * Copyright (c) 2018 Shanghai Zhaoxin Semiconductor Co., Ltd.
 * All Rights Reserved.
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Shanghai Zhaoxin Semiconductor Co., Ltd.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Shanghai Zhaoxin Semiconductor Co., Ltd.
 *
 * The copyright of the source code is protected by the copyright laws of the People's
 * Republic of China and the related laws promulgated by the People's Republic of China
 * and the international covenant(s) ratified by the People's Republic of China.
 *
 */
#include "zx.h"
#include "zx_debugfs.h"
#include "zx_gem.h"
#include "zx_gem_priv.h"
#include "zx_fence.h"
#include "zx_driver.h"
#include "zxgfx_trace.h"
#include "os_interface.h"

#if DRM_VERSION_CODE >= KERNEL_VERSION(5, 16, 0)
MODULE_IMPORT_NS(DMA_BUF);
#endif

extern struct os_pages_memory* zx_allocate_pages_memory_struct(int page_cnt, struct sg_table *st);
extern void zx_pages_memory_extract_st(struct os_pages_memory *memory);
extern unsigned char zx_validate_page_cache(struct os_pages_memory *memory, int start_page, int end_page, unsigned char request_cache_type);
int zx_gem_debugfs_add_object(struct drm_zx_gem_object *obj);
void zx_gem_debugfs_remove_object(struct drm_zx_gem_object *obj);

#define ZX_DEFAULT_PREFAULT_NUM 16
#define NUM_PAGES(x) (((x) + PAGE_SIZE-1) / PAGE_SIZE)

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
#define DMA_BUF_MAP_CLEAR dma_buf_map_clear(map);
#else
#define DMA_BUF_MAP_CLEAR
#endif

unsigned int zx_get_from_gem_handle(void *file_, unsigned int handle)
{
    struct drm_file *file = file_;
    zx_file_t  *priv = file->driver_priv;

    return zx_gem_get_core_handle(priv, handle);
}

static int zx_gem_object_pin_pages(struct drm_zx_gem_object *obj)
{
    int err = 0;

    mutex_lock(&obj->mm.lock);
    if (++obj->mm.pages_pin_count == 1)
    {
        zx_assert(!obj->mm.pages, ZX_FUNC_NAME(__func__));
        err = obj->ops->get_pages(obj);
        if (err)
            --obj->mm.pages_pin_count;
    }
    mutex_unlock(&obj->mm.lock);
    return err;
}

static void zx_gem_object_unpin_pages(struct drm_zx_gem_object *obj)
{
    struct sg_table *pages;

    mutex_lock(&obj->mm.lock);
    zx_assert(obj->mm.pages_pin_count > 0, ZX_FUNC_NAME(__func__));
    if (--obj->mm.pages_pin_count == 0)
    {
        pages = obj->mm.pages;
        obj->mm.pages = NULL;
        zx_assert(pages != NULL, ZX_FUNC_NAME(__func__));
        if (!IS_ERR(pages))
            obj->ops->put_pages(obj, pages);
    }
    mutex_unlock(&obj->mm.lock);
}

void zx_gem_free_object(struct drm_gem_object *gem_obj)
{
    struct drm_zx_gem_object *obj = to_zx_bo(gem_obj);

    trace_zxgfx_drm_gem_release_object(obj);

    drm_gem_object_release(&obj->base);

    if (obj->ops->release)
    {
        obj->ops->release(obj);
    }

    reservation_object_fini(&obj->__builtin_resv);

    zx_assert(!obj->mm.pages_pin_count, ZX_FUNC_NAME(__func__));
    zx_assert(!obj->mm.pages, ZX_FUNC_NAME(__func__));

    zx_free(obj);
}

static struct sg_table *zx_gem_map_dma_buf(struct dma_buf_attachment *attachment, enum dma_data_direction dir)
{
    struct sg_table *st;
    struct scatterlist *src, *dst;
    struct drm_zx_gem_object *obj = dmabuf_to_zx_bo(attachment->dmabuf);
    zx_card_t *zx = obj->base.dev->dev_private;
    int ret, i;

    zx_core_interface->prepare_and_mark_unpagable(zx->adapter, obj->core_handle, &obj->info);

    ret = zx_gem_object_pin_pages(obj);
    if (ret)
        goto err;

    st = kmalloc(sizeof(struct sg_table), GFP_KERNEL);
    if (st == NULL)
    {
        ret = -ENOMEM;
        goto err_unpin_pages;
    }

    ret = sg_alloc_table(st, obj->mm.pages->nents, GFP_KERNEL);
    if (ret)
        goto err_free;

    src = obj->mm.pages->sgl;
    dst = st->sgl;

    for (i = 0; i < obj->mm.pages->nents; i++)
    {
        sg_set_page(dst, sg_page(src), src->length, 0);
        dst = sg_next(dst);
        src = sg_next(src);
    }

    st->nents = dma_map_sg(attachment->dev, st->sgl, st->orig_nents, dir);
    if (!st->nents)
    {
        ret = -ENOMEM;
        goto err_free_sg;
    }
    return st;

err_free_sg:
    sg_free_table(st);
err_free:
    kfree(st);
err_unpin_pages:
    zx_gem_object_unpin_pages(obj);
err:
    return ERR_PTR(ret);
}

static void zx_gem_unmap_dma_buf(struct dma_buf_attachment *attachment, struct sg_table *sg, enum dma_data_direction dir)
{
    struct drm_zx_gem_object *obj = dmabuf_to_zx_bo(attachment->dmabuf);
    zx_card_t *zx = obj->base.dev->dev_private;

    dma_unmap_sg(attachment->dev, sg->sgl, sg->nents, dir);
    sg_free_table(sg);
    kfree(sg);

    zx_gem_object_unpin_pages(obj);
    zx_core_interface->mark_pagable(zx->adapter, obj->core_handle);
}

void *zx_gem_object_vmap(struct drm_zx_gem_object *obj)
{
    zx_card_t *zx = obj->base.dev->dev_private;
    zx_vm_area_t *vma = NULL;
    zx_map_argu_t map_argu = {0, };

    zx_core_interface->prepare_and_mark_unpagable(zx->adapter, obj->core_handle, &obj->info);

    mutex_lock(&obj->mm.lock);
    if (obj->krnl_vma)
    {
        vma = obj->krnl_vma;
        obj->krnl_vma->ref_cnt++;
        goto unlock;
    }

    map_argu.flags.mem_space = ZX_MEM_KERNEL;
    map_argu.flags.read_only = false;
    zx_core_interface->get_map_allocation_info(zx->adapter, obj->core_handle, &map_argu);

    zx_assert(map_argu.flags.mem_space == ZX_MEM_KERNEL, ZX_FUNC_NAME(__func__));
    if (map_argu.flags.mem_type == ZX_SYSTEM_IO)
    {
        vma = zx_map_io_memory(NULL, &map_argu);
    }
    else if (map_argu.flags.mem_type == ZX_SYSTEM_RAM)
    {
        vma = zx_map_pages_memory(NULL, &map_argu);
    }
    else
    {
        zx_assert(0, ZX_FUNC_NAME(__func__));
    }

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
    obj->vmap_mem_type = map_argu.flags.mem_type;
#endif

    obj->krnl_vma = vma;
unlock:
    mutex_unlock(&obj->mm.lock);
    return vma->virt_addr;
}

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static int zx_gem_object_vmap_wrapper(struct drm_gem_object *gem, struct dma_buf_map *map)
{
    struct drm_zx_gem_object *obj= to_zx_bo(gem);
    void *virt_addr = NULL;

    virt_addr = zx_gem_object_vmap(obj);

    if (!virt_addr)
        return -ENOMEM;

    if (obj->vmap_mem_type == ZX_SYSTEM_IO)
    {
        dma_buf_map_set_vaddr_iomem(map, (void __iomem *)virt_addr);

    }
    else if (obj->vmap_mem_type == ZX_SYSTEM_RAM)
    {
        dma_buf_map_set_vaddr(map, virt_addr);
    }
    else
    {
        return -ENODEV;
    }

    return 0;
}
#endif

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static int zx_gem_dmabuf_vmap(struct dma_buf *dma_buf, struct dma_buf_map *map)
#else
static void *zx_gem_dmabuf_vmap(struct dma_buf *dma_buf)
#endif
{
#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
    struct drm_gem_object *obj = dma_buf->priv;
    int ret;

    if(!obj->funcs->vmap)
        return -EOPNOTSUPP;
    
    ret = obj->funcs->vmap(obj, map);
    if(ret)
        return ret;
    else if(dma_buf_map_is_null(map))
        return -ENOMEM;
	
    return 0;
#else
    struct drm_zx_gem_object *obj= dmabuf_to_zx_bo(dma_buf);

    return zx_gem_object_vmap(obj);
#endif
}

void zx_gem_object_vunmap(struct drm_zx_gem_object *obj)
{
    zx_card_t *zx = obj->base.dev->dev_private;

    mutex_lock(&obj->mm.lock);
    if (obj->krnl_vma && --obj->krnl_vma->ref_cnt == 0)
    {
        switch(obj->krnl_vma->flags.mem_type)
        {
        case ZX_SYSTEM_IO:
            zx_unmap_io_memory(obj->krnl_vma);
            break;
        case ZX_SYSTEM_RAM:
            zx_unmap_pages_memory(obj->krnl_vma);
            break;
        default:
            zx_assert(0, ZX_FUNC_NAME(__func__));
            break;
        }
        obj->krnl_vma = NULL;

        zx_core_interface->mark_pagable(zx->adapter, obj->core_handle);
    }
    mutex_unlock(&obj->mm.lock);
}

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static void zx_gem_object_vunmap_wrapper(struct drm_gem_object *gem, struct dma_buf_map *map)
{
    struct drm_zx_gem_object *obj= to_zx_bo(gem);
    
    zx_gem_object_vunmap(obj);
    
    DMA_BUF_MAP_CLEAR
}
#endif

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static void zx_gem_dmabuf_vunmap(struct dma_buf *dma_buf, struct dma_buf_map *map)
#else
static void zx_gem_dmabuf_vunmap(struct dma_buf *dma_buf, void *vaddr)
#endif
{
#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
    struct drm_gem_object *obj = dma_buf->priv;

    if(!obj->funcs->vunmap)
    {
	zx_error("drm gem object funcs vunmap not register.\n");
	DMA_BUF_MAP_CLEAR
	return;
    }
    else
    {
	obj->funcs->vunmap(obj, map);
    }
#else
    struct drm_zx_gem_object *obj = dmabuf_to_zx_bo(dma_buf);

    zx_gem_object_vunmap(obj);
#endif
}

static void *zx_gem_dmabuf_kmap_atomic(struct dma_buf *dma_buf, unsigned long page_num)
{
    zx_info("kmap_atomic: not support yet.\n");
    return NULL;
}

static void zx_gem_dmabuf_kunmap_atomic(struct dma_buf *dma_buf, unsigned long page_num, void *addr)
{
    zx_info("kunmap_atomic: not support yet.\n");
}

static void *zx_gem_dmabuf_kmap(struct dma_buf *dma_buf, unsigned long page_num)
{
    zx_info("kmap: not support yet.\n");
    return NULL;
}

static void zx_gem_dmabuf_kunmap(struct dma_buf *dma_buf, unsigned long page_num, void *addr)
{
    zx_info("kunmap: not support yet.\n");
}

static int zx_gem_dmabuf_mmap(struct dma_buf *dma_buf, struct vm_area_struct *vma)
{
    struct drm_zx_gem_object *obj = dmabuf_to_zx_bo(dma_buf);
    zx_card_t *zx = obj->base.dev->dev_private;
    zx_map_argu_t map_argu = {0, };
    int ret = 0;
     
    zx_info("zx_gem_dmabuf_mmap(%p)\n", dma_buf);

    mutex_lock(&obj->mm.lock);

    map_argu.flags.mem_space = ZX_MEM_USER;
    map_argu.flags.read_only = false;
    zx_core_interface->get_map_allocation_info(zx->adapter, obj->core_handle, &map_argu);

    zx_assert(map_argu.flags.mem_space == ZX_MEM_USER, ZX_FUNC_NAME(__func__));

    switch(map_argu.flags.mem_type)
    {
    case ZX_SYSTEM_IO:
        vma->vm_pgoff = map_argu.phys_addr >> PAGE_SHIFT;
        ret = zx_map_system_io(vma, &map_argu);
        break;
    case ZX_SYSTEM_RAM:
        {
            int start_page = _ALIGN_DOWN(map_argu.offset, PAGE_SIZE)/PAGE_SIZE;
            int end_page = start_page + ALIGN(map_argu.size, PAGE_SIZE) / PAGE_SIZE;
            map_argu.flags.cache_type = zx_validate_page_cache(map_argu.memory, start_page, end_page, map_argu.flags.cache_type);
            ret = zx_map_system_ram(vma, &map_argu);
            break;
        }
    default:
        zx_assert(0, ZX_FUNC_NAME(__func__));
        break;
    }

    mutex_unlock(&obj->mm.lock);
    return ret;
}


#if DRM_VERSION_CODE < KERNEL_VERSION(4,6,0) || (defined(YHQILIN) && DRM_VERSION_CODE == KERNEL_VERSION(4,9,0))
static int zx_gem_begin_cpu_access(struct dma_buf *dma_buf, size_t s1, size_t s2, enum dma_data_direction direction)
#else
static int zx_gem_begin_cpu_access(struct dma_buf *dma_buf, enum dma_data_direction direction)
#endif
{
    struct drm_zx_gem_object *obj= dmabuf_to_zx_bo(dma_buf);
    int write = (direction == DMA_BIDIRECTIONAL || direction == DMA_TO_DEVICE);

    zx_gem_object_begin_cpu_access(obj, BEGIN_CPU_ACCESS_TIMEOUT, write);

    return 0;
}

#if DRM_VERSION_CODE < KERNEL_VERSION(4,6,0) || defined(YHQILIN)
static void zx_gem_end_cpu_access(struct dma_buf *dma_buf, size_t s1, size_t s2, enum dma_data_direction direction)
#else
static int zx_gem_end_cpu_access(struct dma_buf *dma_buf, enum dma_data_direction direction)
#endif
{
    struct drm_zx_gem_object *obj = dmabuf_to_zx_bo(dma_buf);
    int write = (direction == DMA_BIDIRECTIONAL || direction == DMA_TO_DEVICE);

    zx_gem_object_end_cpu_access(obj, write);
#if DRM_VERSION_CODE >= KERNEL_VERSION(4,6,0) && !defined(YHQILIN)
    return 0;
#endif
}

static struct os_pages_memory* zx_allocate_pages_memory_from_sg(struct sg_table *st, int size)
{
    int page_cnt = ALIGN(size, PAGE_SIZE) / PAGE_SIZE;
    struct os_pages_memory *memory = NULL;

    memory = zx_allocate_pages_memory_struct(page_cnt, st);
    if (!memory)
        return NULL;

    memory->shared              = TRUE;
    memory->size                = ALIGN(size, PAGE_SIZE);
    memory->need_flush          = TRUE;
    memory->need_zero           = FALSE;
    memory->fixed_page          = TRUE;
    memory->page_size           = PAGE_SIZE;
    memory->dma32               = FALSE;
    memory->page_4K_64K         = FALSE;
    memory->st                  = st;
    memory->has_dma_map         = TRUE;
    zx_pages_memory_extract_st(memory);
    return memory;
}

static int zx_gem_object_get_pages_generic(struct drm_zx_gem_object *obj)
{
    struct os_pages_memory *memory = NULL;
    zx_card_t *zx = obj->base.dev->dev_private;

    memory = zx_core_interface->get_allocation_pages(zx->adapter, obj->core_handle);
    zx_assert(memory != NULL, ZX_FUNC_NAME(__func__));
    obj->mm.pages = memory->st;

    return 0;
}

static void zx_gem_object_put_pages_generic(struct drm_zx_gem_object *obj, struct sg_table *pages)
{
}

static void zx_gem_object_release_generic(struct drm_zx_gem_object *obj)
{
    zx_destroy_allocation_t destroy = {0, };
    zx_card_t *zx = obj->base.dev->dev_private;

    destroy.device = 0;
    destroy.allocation = obj->core_handle;
    if (zx->a_info.debugfs_mask & ZX_DEBUGFS_GEM_ENABLE)
    {
        zx_gem_debugfs_remove_object(obj);
    }
    zx_core_interface->destroy_allocation(zx->adapter, &destroy);
    obj->core_handle = 0;
}


static const struct drm_zx_gem_object_ops zx_gem_object_generic_ops =
{
    .get_pages = zx_gem_object_get_pages_generic,
    .put_pages = zx_gem_object_put_pages_generic,
    .release = zx_gem_object_release_generic,
};

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static const struct drm_gem_object_funcs zx_gem_object_funcs =
{
    //.close         = ,
    //.open          = ,
    .free          = zx_gem_free_object,
    //mmp is still in drm_driver
    //.mmp
    .vm_ops        = &zx_gem_vm_ops,
    //.pin           = ,
    //.unpin         = ,
    .vmap          = zx_gem_object_vmap_wrapper,
    .vunmap        = zx_gem_object_vunmap_wrapper,
    .export        = zx_gem_prime_export,
    //.get_set_table = ,
};
#endif

static int zx_drm_gem_object_mmap_immediate(struct vm_area_struct* vma, zx_map_argu_t *map_argu)
{
    int ret = 0;

    switch(map_argu->flags.mem_type)
    {
    case ZX_SYSTEM_IO:
        vma->vm_pgoff = map_argu->phys_addr >> PAGE_SHIFT;
#if defined (__aarch64__) || defined (__mips64__)
         map_argu->flags.cache_type = ZX_MEM_UNCACHED;
#endif
        ret = zx_map_system_io(vma, map_argu);
        break;
    case ZX_SYSTEM_RAM:
        {
            int start_page = _ALIGN_DOWN(map_argu->offset, PAGE_SIZE)/PAGE_SIZE;
            int end_page = start_page + ALIGN(map_argu->size, PAGE_SIZE) / PAGE_SIZE;
            map_argu->flags.cache_type = zx_validate_page_cache(map_argu->memory, start_page, end_page, map_argu->flags.cache_type);
            ret = zx_map_system_ram(vma, map_argu);
            break;
        }
    default:
        zx_assert(0, ZX_FUNC_NAME(__func__));
        break;
    }

    return ret;
}

static void zx_drm_gem_object_mmap_delay(struct vm_area_struct* vma, zx_map_argu_t *map_argu)
{
    unsigned int  cache_type;

    switch(map_argu->flags.mem_type)
    {
    case ZX_SYSTEM_IO:
        cache_type = map_argu->flags.cache_type;
#if defined (__aarch64__) || defined (__mips64__)
        cache_type = ZX_MEM_UNCACHED;
#endif
#if LINUX_VERSION_CODE >= 0x020612
        vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
        vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 1);
        map_argu->flags.cache_type = cache_type;

        break;
    case ZX_SYSTEM_RAM:
        // actually map 0 pages here, just force update cache_type
        map_argu->flags.cache_type = zx_validate_page_cache(map_argu->memory, 0, 0, map_argu->flags.cache_type);
        cache_type = map_argu->flags.cache_type;

#if LINUX_VERSION_CODE >= 0x020612
        vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
        vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 0);

        break;
    default:
        zx_assert(0, ZX_FUNC_NAME(__func__));
        break;
    }
}

static int zx_drm_gem_object_mmap(struct file *filp, struct drm_zx_gem_object *obj, struct vm_area_struct *vma)
{
    struct drm_file *file = filp->private_data;
    zx_file_t     *priv = file->driver_priv;
    zx_map_argu_t map_argu = {0, };
    zx_wait_allocation_idle_t wait = {0, };
    unsigned int  cache_type;
    int ret = 0;
    zx_card_t *zx_card = priv->card;
    int i = 0;

    mutex_lock(&obj->mm.lock);

    map_argu.flags.mem_space = ZX_MEM_USER;
    map_argu.flags.read_only = false;
    zx_core_interface->get_map_allocation_info(priv->card->adapter, obj->core_handle, &map_argu);

    wait.engine_mask = ALL_ENGINE_MASK;
    wait.hAllocation = obj->core_handle;
    zx_core_interface->wait_allocation_idle(priv->card->adapter, &wait);

    vma->vm_flags |= (VM_IO | VM_PFNMAP | VM_DONTEXPAND | VM_DONTDUMP);

    if ((map_argu.flags.mem_type == ZX_SYSTEM_RAM) &&
        ((obj->info.segment_id) != 2 && (obj->info.segment_id != 3)))
    {
        zx_error("zx_drm_gem_object_mmap failed: mem_type is ram but segment_id is %d.\n", obj->info.segment_id);
        zx_assert(0, ZX_FUNC_NAME(__func__));
    }

    for (i = 0; i < MAX_CORE_CRTCS; i++)
    {
        if (zx_card->primary_addr[i] == obj->info.gpu_virt_addr)
        {
            //zx_info("trying to mmap primary!!!\n");
            break;
        }
    }

    if (obj->delay_map == 0)
        ret = zx_drm_gem_object_mmap_immediate(vma, &map_argu);
    else
        zx_drm_gem_object_mmap_delay(vma, &map_argu);

    obj->map_argu = map_argu;

    mutex_unlock(&obj->mm.lock);

    return ret;
}


static void zx_drm_gem_object_vm_prepare(struct drm_zx_gem_object *obj)
{
    zx_card_t *zx = obj->base.dev->dev_private;

    zx_core_interface->prepare_and_mark_unpagable(zx->adapter, obj->core_handle, &(obj->info));
}

static void zx_drm_gem_object_vm_release(struct drm_zx_gem_object *obj)
{
    zx_card_t *zx = obj->base.dev->dev_private;

    zx_core_interface->mark_pagable(zx->adapter, obj->core_handle);
}

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
static void zx_gem_object_init(struct drm_zx_gem_object *obj, const struct drm_zx_gem_object_ops *ops, const struct drm_gem_object_funcs *funcs)
#else
static void zx_gem_object_init(struct drm_zx_gem_object *obj, const struct drm_zx_gem_object_ops *ops)
#endif
{
    mutex_init(&obj->mm.lock);
    obj->ops = ops;

    reservation_object_init(&obj->__builtin_resv);
    obj->resv = &obj->__builtin_resv;

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
    obj->base.funcs = funcs;
#endif
}

struct drm_zx_gem_object* zx_drm_gem_create_object(zx_card_t *zx, zx_create_allocation_t *create, zx_device_debug_info_t **ddbg)
{
    struct drm_zx_gem_object *obj = NULL;
    zx_query_info_t query = {0, };
    int result = 0;

    obj = zx_calloc(sizeof(*obj));

    if(create->user_ptr)
    {
        struct os_pages_memory *memory=NULL;

        result = zx_get_pages_from_userptr(&zx->pdev, (unsigned long)create->user_ptr, create->user_buf_size, &memory);
        if(result < 0)
        {
            zx_free(obj);
            return NULL;
        }

        zx_core_interface->create_allocation_from_pages(zx->adapter, create, memory, obj);
    }
    else
    {
        zx_core_interface->create_allocation(zx->adapter, create, obj);
    }

    if(!create->allocation)
    {
    	zx_free(obj);
	    return NULL;
    }

    query.type = ZX_QUERY_ALLOCATION_INFO_KMD;
    query.argu = create->allocation;
    query.buf = &obj->info;
    zx_core_interface->query_info(zx->adapter, &query);
    obj->core_handle = create->allocation;

    drm_gem_private_object_init(_to_drm_device(zx->pdev), &obj->base, create->size);

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
    zx_gem_object_init(obj, &zx_gem_object_generic_ops, &zx_gem_object_funcs);
#else
    zx_gem_object_init(obj, &zx_gem_object_generic_ops);
#endif

    if (zx->a_info.debugfs_mask & ZX_DEBUGFS_GEM_ENABLE)
    {
        obj->debug.parent_dev = ddbg;
        obj->debug.parent_gem = obj;
        obj->debug.root       = zx_debugfs_get_allocation_root(zx->debugfs_dev);
        obj->debug.is_dma_buf_import = false;

        zx_gem_debugfs_add_object(obj);
    }
    trace_zxgfx_drm_gem_create_object(obj);
    return obj;
}

int zx_drm_gem_create_object_ioctl(struct drm_file *file, zx_create_allocation_t *create)
{
    int ret = 0;
    struct drm_zx_gem_object *obj = NULL;
    zx_file_t *priv = file->driver_priv;
    zx_card_t *zx  = priv->card;

    zx_assert(create->device == priv->gpu_device, ZX_FUNC_NAME(__func__));

    obj = zx_drm_gem_create_object(zx, create, &priv->debug);

    if (!obj)
        return -ENOMEM;

    ret = drm_gem_handle_create(file, &obj->base, &create->allocation);

    zx_assert(ret == 0, ZX_FUNC_NAME(__func__));

    zx_gem_object_put(obj);

    return 0;
}

int zx_drm_gem_create_resource_ioctl(struct drm_file *file, zx_create_resource_t *create)
{
    int ret = 0, i;
    zx_file_t *priv = file->driver_priv;
    zx_card_t *zx  = priv->card;
    int obj_count = create->NumAllocations;
    zx_create_alloc_info_t stack_kinfos[8];
    zx_create_alloc_info_t *kinfo = stack_kinfos;
    zx_create_alloc_info_t __user *uinfo = ptr64_to_ptr(create->pAllocationInfo);
    struct drm_zx_gem_object *stack_objs[8];
    struct drm_zx_gem_object **objs = stack_objs;
    zx_query_info_t query = {0, };

    zx_assert(create->device == priv->gpu_device, ZX_FUNC_NAME(__func__));
    zx_assert(uinfo != NULL, ZX_FUNC_NAME(__func__));

    if (obj_count > sizeof(stack_kinfos) / sizeof(stack_kinfos[0]))
    {
        kinfo = zx_malloc(obj_count * sizeof(zx_create_alloc_info_t));
    }
    create->pAllocationInfo = ptr_to_ptr64(kinfo);

    if (obj_count > sizeof(stack_objs) / sizeof(stack_objs[0]))
    {
        objs = zx_malloc(obj_count * sizeof(*objs));
    }

    zx_copy_from_user(kinfo, uinfo, obj_count * sizeof(zx_create_alloc_info_t));

    for (i = 0; i < obj_count; i++)
    {
        objs[i] = zx_calloc(sizeof(struct drm_zx_gem_object));

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
        zx_gem_object_init(objs[i], &zx_gem_object_generic_ops, &zx_gem_object_funcs);
#else
        zx_gem_object_init(objs[i], &zx_gem_object_generic_ops);
#endif

    }

    ret = zx_core_interface->create_allocation_list(zx->adapter, create, (void**)objs);
    zx_assert(ret == 0, ZX_FUNC_NAME(__func__));

    create->pAllocationInfo = ptr_to_ptr64(uinfo);
    if (ret)
        goto failed_free;
 
    for (i = 0; i < obj_count; i++)
    {
        objs[i]->core_handle = kinfo[i].hAllocation;
        zx_assert(kinfo[i].hAllocation, ZX_FUNC_NAME(__func__));

        query.type = ZX_QUERY_ALLOCATION_INFO_KMD;
        query.argu = kinfo[i].hAllocation;
        query.buf = &objs[i]->info;
        zx_core_interface->query_info(zx->adapter, &query);

        drm_gem_private_object_init(_to_drm_device(zx->pdev), &objs[i]->base, kinfo[i].Size);
        ret = drm_gem_handle_create(file, &objs[i]->base, &kinfo[i].hAllocation);
        zx_assert(0 == ret, ZX_FUNC_NAME(__func__));

        if (zx->a_info.debugfs_mask & ZX_DEBUGFS_GEM_ENABLE)
        {
            objs[i]->debug.parent_dev = &priv->debug;
            objs[i]->debug.parent_gem = objs[i];
            objs[i]->debug.root       = zx_debugfs_get_allocation_root(zx->debugfs_dev);
            objs[i]->debug.is_dma_buf_import = false;

            zx_gem_debugfs_add_object(objs[i]);
        }
        zx_gem_object_put(objs[i]);
    }

    trace_zxgfx_drm_gem_create_resource(obj_count, objs);
    zx_copy_to_user(uinfo, kinfo, obj_count * sizeof(zx_create_alloc_info_t));

failed_free:
    if (kinfo != stack_kinfos)
    {
        zx_free(kinfo);
    }
    if (objs != stack_objs)
    {
        zx_free(objs);
    }

    return ret;
}


// DMABUF
static int zx_gem_object_get_pages_dmabuf(struct drm_zx_gem_object *obj)
{
    struct sg_table *pages;
    pages = dma_buf_map_attachment(obj->base.import_attach, DMA_BIDIRECTIONAL);
    if (IS_ERR(pages))
        return PTR_ERR(pages);

    obj->mm.pages = pages;
    return 0;
}

static void zx_gem_object_put_pages_dmabuf(struct drm_zx_gem_object *obj, struct sg_table *pages)
{
    dma_buf_unmap_attachment(obj->base.import_attach, pages, DMA_BIDIRECTIONAL);
}

static void zx_gem_object_release_dmabuf(struct drm_zx_gem_object *obj)
{
    zx_destroy_allocation_t destroy = {0, };
    zx_card_t *zx = obj->base.dev->dev_private;
    zx_wait_allocation_idle_t wait = {0, };
    struct dma_buf_attachment *attach;
    struct dma_buf *dma_buf;

    wait.engine_mask = ALL_ENGINE_MASK;
    wait.hAllocation = obj->core_handle;
    zx_core_interface->wait_allocation_idle(zx->adapter, &wait);

    destroy.device = 0;
    destroy.allocation = obj->core_handle;
    zx_core_interface->destroy_allocation(zx->adapter, &destroy);
    obj->core_handle = 0;

    if (zx->a_info.debugfs_mask & ZX_DEBUGFS_GEM_ENABLE)
    {
        zx_gem_debugfs_remove_object(obj);
    }

    zx_gem_object_unpin_pages(obj);
    attach = obj->base.import_attach;
    dma_buf = attach->dmabuf;
    dma_buf_detach(dma_buf, attach);
    dma_buf_put(dma_buf);
}

static const struct drm_zx_gem_object_ops zx_gem_object_dmabuf_ops =
{
    .get_pages = zx_gem_object_get_pages_dmabuf,
    .put_pages = zx_gem_object_put_pages_dmabuf,
    .release = zx_gem_object_release_dmabuf,
};


int zx_gem_prime_fd_to_handle(struct drm_device *dev, struct drm_file *file_priv, int prime_fd, uint32_t *handle)
{
    int ret;
    struct drm_zx_driver *driver = to_drm_zx_driver(dev->driver);

    mutex_lock(&driver->lock);
    zx_assert(!driver->file_priv, ZX_FUNC_NAME(__func__));

    driver->file_priv = file_priv;
    ret = drm_gem_prime_fd_to_handle(dev, file_priv, prime_fd, handle);
    driver->file_priv = NULL;
    mutex_unlock(&driver->lock);

    return ret;
}

signed long zx_gem_object_begin_cpu_access(struct drm_zx_gem_object *obj, long timeout, int write)
{
    trace_zxgfx_gem_object_begin_cpu_access(obj, timeout, write);

    return zx_gem_fence_await_reservation(obj->resv, 0, timeout, write);
}

int zx_gem_object_begin_cpu_access_ioctl(struct drm_file *file, zx_drm_gem_begin_cpu_access_t *args)
{
    int ret = 0;
    struct drm_device *dev = file->minor->dev;
    struct drm_zx_gem_object *obj;

    obj = zx_drm_gem_object_lookup(dev, file, args->handle);

    if (!obj)
        return -ENOENT;

    ret = zx_gem_object_begin_cpu_access(obj, BEGIN_CPU_ACCESS_TIMEOUT, !args->readonly);

    zx_gem_object_put(obj);
    return 0;
}

void zx_gem_object_end_cpu_access(struct drm_zx_gem_object *obj, int write)
{
    zx_card_t *zx = obj->base.dev->dev_private;
    zx_map_argu_t map_argu = {0, };

    map_argu.flags.mem_space = ZX_MEM_KERNEL;
    map_argu.flags.read_only = false;
    zx_core_interface->get_map_allocation_info(zx->adapter, obj->core_handle, &map_argu);

    if (map_argu.flags.mem_type == ZX_SYSTEM_RAM && 
        map_argu.flags.cache_type == ZX_MEM_WRITE_BACK &&
        (!obj->info.snoop) &&
        map_argu.memory)
    {
        zx_flush_cache(zx->pdev, NULL, map_argu.memory, map_argu.offset, map_argu.size);
    }

    if (map_argu.flags.cache_type == ZX_MEM_WRITE_COMBINED)
    {
        zx_wmb();
    }

    trace_zxgfx_gem_object_end_cpu_access(obj);
}

void zx_gem_object_end_cpu_access_ioctl(struct drm_file *file, zx_drm_gem_end_cpu_access_t *args)
{
    struct drm_device *dev = file->minor->dev;
    struct drm_zx_gem_object *obj;

    obj = zx_drm_gem_object_lookup(dev, file, args->handle);

    if (!obj)
        return;

    zx_gem_object_end_cpu_access(obj, 1);
    zx_gem_object_put(obj);
}

const struct dma_buf_ops CONCAT(zx_dmabuf_ops,DRIVER_NAME) =
{
    .map_dma_buf    = zx_gem_map_dma_buf,
    .unmap_dma_buf  = zx_gem_unmap_dma_buf,
    .release        = drm_gem_dmabuf_release,
#if DRM_VERSION_CODE < KERNEL_VERSION(5,6,0)
#if DRM_VERSION_CODE > KERNEL_VERSION(4,11,0)
    .map           = zx_gem_dmabuf_kmap,
    .unmap         = zx_gem_dmabuf_kunmap,
#if DRM_VERSION_CODE < KERNEL_VERSION(4,19,0)
    .map_atomic    = zx_gem_dmabuf_kmap_atomic,
    .unmap_atomic  = zx_gem_dmabuf_kunmap_atomic,
#endif
#else
    .kmap            = zx_gem_dmabuf_kmap,
    .kmap_atomic     = zx_gem_dmabuf_kmap_atomic,
    .kunmap          = zx_gem_dmabuf_kunmap,
    .kunmap_atomic   = zx_gem_dmabuf_kunmap_atomic,
#endif
#endif
    .mmap           = zx_gem_dmabuf_mmap,
    .vmap           = zx_gem_dmabuf_vmap,
    .vunmap         = zx_gem_dmabuf_vunmap,
    .begin_cpu_access   = zx_gem_begin_cpu_access,
    .end_cpu_access     = zx_gem_end_cpu_access,
};

EXPORT_SYMBOL(CONCAT(zx_dmabuf_ops,DRIVER_NAME));

struct drm_gem_object *zx_gem_prime_import(struct drm_device *dev, struct dma_buf *dma_buf)
{
    struct drm_zx_driver *driver = to_drm_zx_driver(dev->driver);
    struct drm_file *drm_file = driver->file_priv;
    zx_file_t *priv = drm_file->driver_priv;
    zx_card_t *zx = dev->dev_private;

    struct drm_zx_gem_object *obj = NULL;
    struct dma_buf_attachment *attach;
    struct os_pages_memory *pages = NULL;
    zx_create_allocation_t create = {0, };
    zx_query_info_t query = {0, };
    int ret;

    zx_assert(priv != NULL, ZX_FUNC_NAME(__func__));
    zx_assert(priv->gpu_device, ZX_FUNC_NAME(__func__));

    if (dma_buf->ops == &CONCAT(zx_dmabuf_ops,DRIVER_NAME))
    {
        obj = dmabuf_to_zx_bo(dma_buf);

        if (obj->base.dev == dev)
        {
            trace_zxgfx_gem_prime_import(dma_buf, obj);
            return &(zx_gem_object_get(obj)->base);
        }
    }

    attach = dma_buf_attach(dma_buf, dev->dev);
    if (IS_ERR(attach))
        return ERR_CAST(attach);

    get_dma_buf(dma_buf);
    obj = zx_calloc(sizeof(*obj));
    if (obj == NULL)
    {
        ret = -ENOMEM;
        goto fail_detach;
    }

    drm_gem_private_object_init(dev, &obj->base, dma_buf->size);

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,11,0)
    zx_gem_object_init(obj, &zx_gem_object_dmabuf_ops, &zx_gem_object_funcs);
#else
    zx_gem_object_init(obj, &zx_gem_object_dmabuf_ops);
#endif

    obj->base.import_attach = attach;
    obj->resv = dma_buf->resv;

    ret = zx_gem_object_pin_pages(obj);
    if (ret != 0)
        goto fail_free;

    pages = zx_allocate_pages_memory_from_sg(obj->mm.pages, dma_buf->size);
    if (!pages)
        goto fail_unpin;

    create.device       = priv->gpu_device;
    create.width        = dma_buf->size;
    create.height       = 1;
    create.usage_mask   = ZX_USAGE_TEMP_BUFFER;
    create.format       = ZX_FORMAT_A8_UNORM;
    create.access_hint  = ZX_ACCESS_CPU_ALMOST;
    
    zx_core_interface->create_allocation_from_pages(priv->card->adapter, &create, pages, obj);
    zx_assert(create.allocation != 0, ZX_FUNC_NAME(__func__));
    obj->core_handle = create.allocation;

    if (!obj->core_handle)
    {
        ret = -ENOMEM;
        goto fail_unpin;
    }

    query.type = ZX_QUERY_ALLOCATION_INFO_KMD;
    query.argu = create.allocation;
    query.buf = &obj->info;
    zx_core_interface->query_info(priv->card->adapter, &query);

    if (zx->a_info.debugfs_mask & ZX_DEBUGFS_GEM_ENABLE)
    {
        obj->debug.parent_gem = obj;
        obj->debug.root       = zx_debugfs_get_allocation_root(zx->debugfs_dev);
        obj->debug.parent_dev = &priv->debug;
        obj->debug.is_dma_buf_import = true;
        zx_gem_debugfs_add_object(obj);
    }
    trace_zxgfx_gem_prime_import(dma_buf, obj);
    return &obj->base;

fail_unpin:
    if (pages)
        zx_free_pages_memory(to_pci_dev(dev->dev), pages);
    zx_gem_object_unpin_pages(obj);
fail_free:
    zx_free(obj);
fail_detach:
    dma_buf_detach(dma_buf, attach);
    dma_buf_put(dma_buf);

    return ERR_PTR(ret);
}

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,4,0)
struct dma_buf *zx_gem_prime_export(struct drm_gem_object *gem_obj, int flags)
#else
struct dma_buf *zx_gem_prime_export(struct drm_device *dev, struct drm_gem_object *gem_obj, int flags)
#endif
{
    struct dma_buf *dma_buf;
    struct drm_zx_gem_object *obj;
    struct drm_device  *dev1;
    DEFINE_DMA_BUF_EXPORT_INFO(exp_info);

#if DRM_VERSION_CODE >= KERNEL_VERSION(5,4,0)
    dev1 = gem_obj->dev;
#else
    dev1 = dev;
#endif
    obj = to_zx_bo(gem_obj);

    exp_info.ops = &CONCAT(zx_dmabuf_ops,DRIVER_NAME);
    exp_info.size = obj->base.size;
    exp_info.flags = flags;
    exp_info.priv = gem_obj;
    exp_info.resv = obj->resv;

#if DRM_VERSION_CODE < KERNEL_VERSION(4,9,0)
    dma_buf = dma_buf_export(&exp_info);
#else
    dma_buf = drm_gem_dmabuf_export(dev1, &exp_info);
#endif

    trace_zxgfx_gem_prime_export(dma_buf, obj);

    return dma_buf;
}

int zx_gem_mmap_gtt(struct drm_file *file, struct drm_device *dev, uint32_t handle, uint64_t *offset)
{
    int ret;
    struct drm_zx_gem_object *obj;

    obj = zx_drm_gem_object_lookup(dev, file, handle);
    if (!obj)
    {
        zx_error("zx_gem_mmap_gtt lookup failed \n");
        ret = -ENOENT;
        goto unlock;
    }

    ret = drm_gem_create_mmap_offset(&obj->base);
    if (ret)
    {
        zx_error("zx_gem_mmap_gtt create mmap offset failed \n");
        goto out;
    }
    *offset = drm_vma_node_offset_addr(&obj->base.vma_node);

out:
    zx_gem_object_put(obj);
unlock:
    return ret;
}

int zx_gem_mmap_gtt_ioctl(struct drm_file *file, zx_drm_gem_map_t *args)
{
    int ret;
    struct drm_zx_gem_object *obj;

    obj = zx_drm_gem_object_lookup(file->minor->dev, file, args->gem_handle);
    if (!obj)
    {
        zx_error("zx_gem_mmap_gtt_ioctl lookup failed \n");
        ret = -ENOENT;
        goto unlock;
    }

    obj->delay_map = args->delay_map;
    obj->prefault_num = args->prefault_num;

    ret = drm_gem_create_mmap_offset(&obj->base);
    if (ret)
    {
        zx_error("zx_gem_mmap_gtt_ioctl create mmap offset failed \n");
        goto out;
    }
    args->offset = drm_vma_node_offset_addr(&obj->base.vma_node);

out:
    zx_gem_object_put(obj);
unlock:
    return ret;
}

int zx_drm_gem_mmap(struct file *filp, struct vm_area_struct *vma)
{
    struct drm_zx_gem_object *obj;
    struct drm_file *file = filp->private_data;
    zx_file_t     *priv = file->driver_priv;
    int             ret = 0;

    if (priv->map)
    {
        return zx_mmap(filp, vma);
    }

    ret = drm_gem_mmap(filp, vma);
    if (ret < 0)
    {
        zx_error("failed to mmap, ret = %d.\n", ret);
        return ret;
    }

    obj = to_zx_bo(vma->vm_private_data);

    vma->vm_flags &= ~(VM_IO | VM_PFNMAP | VM_DONTEXPAND | VM_DONTDUMP);
    vma->vm_page_prot = pgprot_writecombine(vm_get_page_prot(vma->vm_flags));

    zx_drm_gem_object_vm_prepare(obj);

    return zx_drm_gem_object_mmap(filp, obj, vma);
}

int zx_gem_dumb_create(struct drm_file *file, struct drm_device *dev, struct drm_mode_create_dumb *args)
{
    int ret = 0;
    struct drm_zx_gem_object *obj = NULL;
    zx_file_t *priv = file->driver_priv;
    zx_card_t *zx  = NULL;
    zx_create_allocation_t create = {0, };
    zx_format zxFormat = ZX_FORMAT_R8_UNORM;

    if(args->bpp == 32)
    {
        zxFormat = ZX_FORMAT_B8G8R8A8_UNORM;
    }
    else if(args->bpp == 16)
    {
        zxFormat = ZX_FORMAT_B5G6R5_UNORM;
    }

    //
    // In cmodel + hybrid, Xserver will call zx_gem_dumb_create before kmd is really loadded
    // so add following checking
    //
    if(priv == NULL)
    {
        zx_info("%s() zx_file_t is not created yet\n", __func__);
        args->pitch = 0;
        args->size = 0;
        return -1;
    }

    zx = priv->card;

    create.device   = priv->gpu_device;
    create.width    = args->width;
    create.height   = args->height;
    create.format   = zxFormat;
    create.tiled    = 0;
    create.unpagable    = TRUE;
    create.usage_mask   = ZX_USAGE_DISPLAY_SURFACE | ZX_USAGE_FRAMEBUFFER;
    create.access_hint  = ZX_ACCESS_CPU_ALMOST;
    create.primary      = 1;

    obj = zx_drm_gem_create_object(zx, &create, &priv->debug);
    if (!obj)
        return -ENOMEM;

    ret = drm_gem_handle_create(file, &obj->base, &args->handle);
    zx_assert(ret == 0, ZX_FUNC_NAME(__func__));

    args->pitch = create.pitch;
    args->size  = create.size;
    zx_gem_object_put(obj);

    return ret;
}

static vm_fault_t zx_gem_io_insert(struct vm_area_struct *vma, unsigned long address, zx_map_argu_t *map, unsigned int offset, unsigned int prefault_num)
{
    unsigned long pfn;
    vm_fault_t retval = VM_FAULT_NOPAGE;
    int i ;

    for(i = 0; i < prefault_num ; i++)
    {
        if(offset >= map->size)
            break;

        pfn = (map->phys_addr + offset) >> PAGE_SHIFT;

        retval = vmf_insert_pfn(vma, address, pfn);
        if (unlikely((retval == VM_FAULT_NOPAGE && i > 0)))
            break;
        else if (unlikely(retval & VM_FAULT_ERROR))
        {
            return retval;
        }

        offset += PAGE_SIZE;
        address += PAGE_SIZE;
    }

    return retval;
}

static vm_fault_t zx_gem_ram_insert(struct vm_area_struct *vma, unsigned long address, zx_map_argu_t *map, unsigned int offset, unsigned int prefault_num)
{
    unsigned long pfn;
    vm_fault_t retval = VM_FAULT_NOPAGE;
    int i, start_page, end_page;

    start_page = _ALIGN_DOWN(offset, PAGE_SIZE) / PAGE_SIZE;
    end_page = min((int)(start_page+prefault_num), (int)(map->memory->size / PAGE_SIZE));
    map->flags.cache_type = zx_validate_page_cache(map->memory, start_page, end_page, map->flags.cache_type);

    for (i = start_page; i < end_page; i++)
    {
        pfn  = page_to_pfn(map->memory->pages[i]);
        retval = vmf_insert_pfn(vma, address, pfn);
        if (unlikely((retval == VM_FAULT_NOPAGE && i > start_page)))
            break;
        else if (unlikely(retval & VM_FAULT_ERROR))
        {
            return retval;
        }
        address += PAGE_SIZE;
    }
    return retval;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(4,11,0)
static vm_fault_t zx_gem_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
#else
static vm_fault_t zx_gem_fault(struct vm_fault *vmf)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    struct vm_area_struct *vma = vmf->vma;
#endif
    struct drm_zx_gem_object *obj;
    unsigned long offset;
    uint64_t vma_node_offset;
    vm_fault_t ret = VM_FAULT_NOPAGE;
    unsigned long address;
    unsigned int prefault_num = 0;
    unsigned int page_num = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,10,0)
    address = vmf->address;
#else
    address = (unsigned long)vmf->virtual_address;
#endif
    obj = to_zx_bo(vma->vm_private_data);

    vma_node_offset = drm_vma_node_offset_addr(&(obj->base.vma_node));
    offset = ((vma->vm_pgoff << PAGE_SHIFT) - vma_node_offset) + (address - vma->vm_start);

    zx_assert(obj->delay_map != 0, ZX_FUNC_NAME(__func__));

    page_num = NUM_PAGES(vma->vm_end - address);

    if (obj->prefault_num == 0)
        prefault_num = min((unsigned int)ZX_DEFAULT_PREFAULT_NUM, page_num);
    else
        prefault_num = min(obj->prefault_num, page_num);

    switch(obj->map_argu.flags.mem_type)
    {
    case ZX_SYSTEM_IO:
        ret = zx_gem_io_insert(vma, address, &obj->map_argu, offset, prefault_num);
        break;
    case ZX_SYSTEM_RAM:
        ret = zx_gem_ram_insert(vma, address, &obj->map_argu, offset, prefault_num);
        break;
    default:
        zx_assert(0, ZX_FUNC_NAME(__func__));
        break;
    }

    return ret;
}

static void zx_gem_vm_open(struct vm_area_struct *vma)
{
    struct drm_zx_gem_object *obj = to_zx_bo(vma->vm_private_data);

    zx_drm_gem_object_vm_prepare(obj);
    drm_gem_vm_open(vma);
}

static void zx_gem_vm_close(struct vm_area_struct *vma)
{
    struct drm_zx_gem_object *obj = to_zx_bo(vma->vm_private_data);

    zx_drm_gem_object_vm_release(obj);
    drm_gem_vm_close(vma);
}

static int zx_gem_vm_access(struct vm_area_struct *vma, unsigned long addr, void *buf, int len, int write)
{
    void *virtul_addr = NULL;
    struct drm_zx_gem_object *obj;
    unsigned long offset;
    int rest, this_size;
    char *p;

    obj = to_zx_bo(vma->vm_private_data);
    offset = addr - vma->vm_start + ((vma->vm_pgoff - drm_vma_node_start(&obj->base.vma_node)) << PAGE_SHIFT);
    //zx_info("zx_gem_vm_access(addr:%p, len:%d, write:%d, offset:%llx)\n", (void*)addr, len, write, offset);

    if (len < 1 || (offset + len) > obj->base.size)
    {
        zx_info("  access overflow (size:%x, vm_start:%llx, vm_pgoff:%llx, node_offset:%llx)\n",
            obj->base.size, vma->vm_start, vma->vm_pgoff, drm_vma_node_start(&obj->base.vma_node));
        return -EIO;
    }

    virtul_addr = zx_gem_object_vmap(obj);
    if (!virtul_addr)
    {
        zx_info("  vmap failed.\n");
        return -EIO;
    }

    rest = len;
    p = (char*)buf;


#define __MIN(a,b)  ((a) > (b) ? (b) : (a))

    this_size = __MIN(rest, obj->base.size - offset);
    if (write)
        zx_memcpy(virtul_addr + offset, p, this_size);
    else
        zx_memcpy(p, virtul_addr + offset, this_size);
    rest -= this_size;

    zx_gem_object_vunmap(obj);

    return len - rest;
}

const struct vm_operations_struct zx_gem_vm_ops = {
    .fault = zx_gem_fault,
    .open = zx_gem_vm_open,
    .close = zx_gem_vm_close,
    .access = zx_gem_vm_access,
};


static int zx_gem_debugfs_show_info(struct seq_file *s, void *unused)
{
    struct zx_gem_debug_info *dnode = (struct zx_gem_debug_info *)(s->private);
    struct drm_zx_gem_object *obj = (struct drm_zx_gem_object *) dnode->parent_gem;
    zx_open_allocation_t *info = &obj->info;
/*print the info*/
    seq_printf(s, "device=0x%x\n", info->device);
    seq_printf(s, "handle=0x%x\n", info->allocation);

    seq_printf(s, "width=%d\n", info->width);
    seq_printf(s, "height=%d\n", info->height);
    seq_printf(s, "pitch=%d\n", info->pitch);
    seq_printf(s, "bit_cnt=%d\n", info->bit_cnt);

    seq_printf(s, "secured=%d\n", info->secured);
    seq_printf(s, "size=%d\n", info->size);
    seq_printf(s, "tile=%d\n", info->tiled);
    seq_printf(s, "segmentid=%d\n", info->segment_id);
    seq_printf(s, "hw_format=%d\n", info->hw_format);
    seq_printf(s, "compress_format=%d\n", info->compress_format);
    seq_printf(s, "bl_slot_index=%d\n", info->bl_slot_index);
    seq_printf(s, "gpu_virt_addr=%lld\n", info->gpu_virt_addr);
    seq_printf(s, "imported=%d\n", dnode->is_dma_buf_import);

    return 0;
}

static int zx_gem_debugfs_info_open(struct inode *inode, struct file *file)
{
    return single_open(file, zx_gem_debugfs_show_info, inode->i_private);
}

static const struct file_operations debugfs_gem_info_fops = {
        .open       = zx_gem_debugfs_info_open,
        .read       = seq_read,
        .llseek     = seq_lseek,
        .release    = single_release,
};

static ssize_t zx_gem_debugfs_data_read(struct file *f, char __user *buf,
                     size_t size, loff_t *pos)
{
  //map & read
    zx_gem_debug_info_t *dbg = file_inode(f)->i_private;
    struct drm_zx_gem_object *obj = dbg->parent_gem;
    void *data = NULL;
    ssize_t result = 0;
    int data_off = 0;
    int copy_size = 0;
    int a_size = 0;

    data = zx_gem_object_vmap(obj);
    if (data) {
        zx_gem_object_begin_cpu_access(obj, BEGIN_CPU_ACCESS_TIMEOUT, 0);
        a_size = obj->krnl_vma->size;
        data_off = *pos;
        copy_size = min(size, (size_t)(a_size - data_off));
        if (copy_to_user(buf,  data + data_off, copy_size)){
            result = -EFAULT;
        } else {
            result += copy_size;
            *pos += copy_size;
        }
        zx_gem_object_end_cpu_access(obj,0);
    }
    else {
        result = -EINVAL;
    }
    zx_gem_object_vunmap(obj);

    return result;
}

static const struct file_operations debugfs_gem_data_fops = {
        .read       = zx_gem_debugfs_data_read,
        .llseek     = default_llseek,
};

static ssize_t zx_gem_debugfs_control_write(struct file *f, const char __user *buf,
                                         size_t size, loff_t *pos)
{
    zx_gem_debug_info_t *dbg = file_inode(f)->i_private;
    struct drm_zx_gem_object *obj = dbg->parent_gem;
    zx_card_t *zx = obj->base.dev->dev_private;
    char c = 0;
    ssize_t ret = 0;

    ret = simple_write_to_buffer(&c, 1, pos, buf, size);
    if (ret > 0) {
        if (c != '0' && !dbg->mark_unpagable) {
            zx_core_interface->prepare_and_mark_unpagable(zx->adapter, obj->core_handle, &obj->info);
            dbg->mark_unpagable = true;
        } else if (c == '0' && dbg->mark_unpagable) {
            zx_core_interface->mark_pagable(zx->adapter, obj->core_handle);
            dbg->mark_unpagable = false;
        }
        ret = size;
    }
    return ret;
}

static ssize_t zx_gem_debugfs_control_read(struct file *f, char __user *buf,
                     size_t size, loff_t *pos)
{
    zx_gem_debug_info_t *dbg = file_inode(f)->i_private;
    struct drm_zx_gem_object *obj = dbg->parent_gem;
    char c = dbg->mark_unpagable ? '1' : '0';

    return simple_read_from_buffer(buf, size, pos, &c, 1);
}

static const struct file_operations debugfs_gem_control_fops = {
        .write       = zx_gem_debugfs_control_write,
        .read       = zx_gem_debugfs_control_read,
        .llseek     = default_llseek,
};

//define from mm_e3k.h
#define BL_BUFFER_SIZE         0x2000000   //32M
#define BL_INDEX_BITS          18
#define BL_MAPPING_RATIO       512

#define BL_SLOT_SIZE           (BL_BUFFER_SIZE >> BL_INDEX_BITS)
#define BL_SLICE_ALIGNMENT     (BL_BUFFER_SIZE >> BL_INDEX_BITS)

static ssize_t zx_gem_debugfs_bl_read(struct file *f, char __user *buf,
                     size_t size, loff_t *pos)
{
    zx_gem_debug_info_t *dbg = file_inode(f)->i_private;
    struct drm_zx_gem_object *obj = dbg->parent_gem;
    zx_card_t *zx = obj->base.dev->dev_private;
    adapter_info_t *ainfo = &zx->adapter_info;

    void *data = NULL;
    ssize_t result = 0;
    int data_off = 0;
    int copy_size = 0;
    int a_size = 0;

    zx_vm_area_t *vma = NULL;
    zx_map_argu_t map_argu = {0, };

    if (!obj->info.bl_slot_index)
    {
        zx_info("uncompress buffer bl_slot_index is 0\n");
        return result;
    }

    map_argu.flags.mem_space = ZX_MEM_KERNEL;
    map_argu.flags.cache_type = ZX_MEM_UNCACHED;
    map_argu.flags.mem_type   = ZX_SYSTEM_IO;
    map_argu.phys_addr = obj->info.bl_slot_index * BL_SLOT_SIZE + ainfo->fb_bus_addr;
    map_argu.size = ALIGN(obj->info.size/BL_MAPPING_RATIO, BL_SLICE_ALIGNMENT); // 1:512, compress ratio

    //zx_info("offset %u, addr %llu size 0x%lu \n",obj->info.bl_slot_index * BL_SLOT_SIZE,  map_argu.phys_addr, map_argu.size);

    vma = zx_map_io_memory(NULL, &map_argu);

    if (vma != NULL)
    {
        data = vma->virt_addr;
        if (data)
        {
            a_size = vma->size;
            data_off = *pos;
            copy_size = min(size, (size_t)(a_size - data_off));
            if (copy_to_user(buf,  data + data_off, copy_size))
            {
               result = -EFAULT;
            }
            else
            {
                result += copy_size;
                *pos += copy_size;
            }
        }
        else
        {
            result = -EINVAL;
        }

        zx_unmap_io_memory(vma);
    }
    else
    {
        zx_error("map bl buffer fail \n");
        result = -EINVAL;
    }

    return result;
}

static const struct file_operations debugfs_gem_bl_fops = {
        .read       = zx_gem_debugfs_bl_read,
        .llseek     = default_llseek,
};

int zx_gem_debugfs_add_object(struct drm_zx_gem_object *obj)
{
    zx_gem_debug_info_t *dbg = &obj->debug;
    int result = 0;

    dbg->is_cpu_accessable = true; //s houdle false for cpu invisable

    zx_vsprintf(dbg->name,"%08x", obj->info.allocation);

    dbg->self_dir = debugfs_create_dir(dbg->name, dbg->root);
    if (dbg->self_dir) {
        dbg->alloc_info = debugfs_create_file("info", 0444, dbg->self_dir, dbg, &debugfs_gem_info_fops);
        dbg->data = debugfs_create_file("data", 0444, dbg->self_dir, dbg, &debugfs_gem_data_fops);
        dbg->control = debugfs_create_file("control", 0444, dbg->self_dir, dbg, &debugfs_gem_control_fops);
        dbg->bl = debugfs_create_file("bl", 0444, dbg->self_dir, dbg, &debugfs_gem_bl_fops);
    }else {
        zx_error("create allocation %s dir failed\n", dbg->name);
        result = -1;
    }
    return result;
}

void zx_gem_debugfs_remove_object(struct drm_zx_gem_object *obj)
{
    zx_gem_debug_info_t *dbg = &obj->debug;
    if (dbg->self_dir) {

        debugfs_remove(dbg->alloc_info);
        debugfs_remove(dbg->data);
        debugfs_remove(dbg->control);
        debugfs_remove(dbg->bl);
        debugfs_remove(dbg->self_dir);

        dbg->alloc_info = NULL;
        dbg->data = NULL;
        dbg->self_dir = NULL;
        dbg->control = NULL;
        dbg->bl = NULL;
    }
}

static inline zx_card_t * get_zxcard(void *dev)
{
    struct drm_device *drm_dev = _to_drm_device(dev);
    return (zx_card_t *)drm_dev->dev_private;
}
struct drm_zx_gem_object* CONCAT(zx_krnl_gem_create_object,DRIVER_NAME)(void *dev,
        zx_create_allocation_t *create, zx_device_debug_info_t **ddev)
{
    zx_card_t *zx = get_zxcard(dev);
    return zx_drm_gem_create_object(zx, create, ddev);
}
EXPORT_SYMBOL(CONCAT(zx_krnl_gem_create_object,DRIVER_NAME));

int CONCAT(zx_krnl_create_device, DRIVER_NAME)(void *dev,
        struct zx_krnl_device_create *vdev_create)
{
    int err = 0;
    zx_card_t *zx = get_zxcard(dev);
    err = zx_core_interface->create_device(zx->adapter, NULL, &vdev_create->gpu_device);
    zx_assert(err == 0, ZX_FUNC_NAME(__func__));
    if(zx->debugfs_dev)
    {
        vdev_create->debug = zx_debugfs_add_device_node(zx->debugfs_dev, zx_get_current_pid(), vdev_create->gpu_device);
    }
    return err;
}

EXPORT_SYMBOL(CONCAT(zx_krnl_create_device,DRIVER_NAME));

void CONCAT(zx_krnl_destroy_device,DRIVER_NAME)(void *dev,
        struct zx_krnl_device_destroy *destroy)
{
    zx_card_t *zx = get_zxcard(dev);
    zx_core_interface->destroy_device(zx->adapter, destroy->gpu_device);
    if (destroy->debug && zx->debugfs_dev) {
        zx_debugfs_remove_device_node(zx->debugfs_dev, destroy->debug);
    }
}

EXPORT_SYMBOL(CONCAT(zx_krnl_destroy_device,DRIVER_NAME));

